import logging
from datetime import timedelta

from core.choices import DataSourceStatusChoices
from core.choices import JobStatusChoices
from core.exceptions import SyncError
from core.models import Job
from rq.timeouts import JobTimeoutException
from utilities.utils import local_now

from .models import IPFabricBranch
from .models import IPFabricSource
from .models import IPFabricSync

logger = logging.getLogger(__name__)


def sync_ipfabricsource(job, *args, **kwargs):
    ipfsource = IPFabricSource.objects.get(pk=job.object_id)

    try:
        job.start()
        ipfsource.sync(job=job)
        job.terminate()
    except Exception as e:
        job.terminate(status=JobStatusChoices.STATUS_ERRORED)
        IPFabricSource.objects.filter(pk=ipfsource.pk).update(
            status=DataSourceStatusChoices.FAILED
        )
        if type(e) in (SyncError, JobTimeoutException):
            logging.error(e)
        else:
            raise e


def sync_ipfabric(job, *args, **kwargs):
    obj = IPFabricSync.objects.get(pk=job.object_id)

    try:
        job.start()
        obj.sync(job=job)
        job.terminate()
    except Exception as e:
        job.terminate(status=JobStatusChoices.STATUS_ERRORED)
        IPFabricSync.objects.filter(pk=obj.pk).update(
            status=DataSourceStatusChoices.FAILED
        )
        if type(e) in (SyncError, JobTimeoutException):
            logging.error(e)
        else:
            raise e
    finally:
        if obj.interval and not kwargs.get("adhoc"):
            new_scheduled_time = local_now() + timedelta(minutes=obj.interval)
            job = Job.enqueue(
                sync_ipfabric,
                name=f"{obj.name} - (scheduled)",
                instance=obj,
                user=obj.user,
                schedule_at=new_scheduled_time,
                interval=obj.interval,
            )


def merge_ipfabric_branch(job, *args, **kwargs):
    branch = IPFabricBranch.objects.get(pk=job.object_id)

    try:
        job.start()
        branch.sync_merge()
        job.terminate()
    except Exception as e:
        job.terminate(status=JobStatusChoices.STATUS_ERRORED)
        IPFabricSync.objects.filter(pk=branch.sync.pk).update(
            status=DataSourceStatusChoices.FAILED
        )
        if type(e) in (SyncError, JobTimeoutException):
            logging.error(e)
        else:
            raise e



from django.urls import include
from django.urls import path
from utilities.urls import get_model_urls

from . import views

urlpatterns = (
    # Source
    path("source/", views.IPFabricSourceListView.as_view(), name="ipfabricsource_list"),
    path(
        "source/add/", views.IPFabricSourceEditView.as_view(), name="ipfabricsource_add"
    ),
    path(
        "source/delete/",
        views.IPFabricSourceBulkDeleteView.as_view(),
        name="ipfabricsource_bulk_delete",
    ),
    path(
        "source/<int:pk>/delete/",
        views.IPFabricSourceDeleteView.as_view(),
        name="ipfabricsource_delete",
    ),
    path(
        "source/<int:pk>/", include(get_model_urls("ipfabric_netbox", "ipfabricsource"))
    ),
    path(
        "source/delete/",
        views.IPFabricSyncBulkDeleteView.as_view(),
        name="ipfabricsync_bulk_delete",
    ),
    # Snapshot
    path(
        "snapshot/",
        views.IPFabricSnapshotListView.as_view(),
        name="ipfabricsnapshot_list",
    ),
    path(
        "snapshot/delete/",
        views.IPFabricSnapshotBulkDeleteView.as_view(),
        name="ipfabricsnapshot_bulk_delete",
    ),
    path(
        "snapshot/<int:pk>/",
        include(get_model_urls("ipfabric_netbox", "ipfabricsnapshot")),
    ),
    path(
        "snapshot/<int:pk>/delete/",
        views.IPFabricSnapshotDeleteView.as_view(),
        name="ipfabricsnapshot_delete",
    ),
    path(
        "data/<int:pk>/delete",
        views.IPFabricSnapshotDataDeleteView.as_view(),
        name="ipfabricdata_delete",
    ),
    path("data/<int:pk>/", include(get_model_urls("ipfabric_netbox", "ipfabricdata"))),
    path(
        "data/delete",
        views.IPFabricSnapshotDataBulkDeleteView.as_view(),
        name="ipfabricdata_bulk_delete",
    ),
    # Sync
    path("sync/", views.IPFabricSyncListView.as_view(), name="ipfabricsync_list"),
    path("sync/add/", views.IPFabricSyncEditView.as_view(), name="ipfabricsync_add"),
    path("sync/<int:pk>/", include(get_model_urls("ipfabric_netbox", "ipfabricsync"))),
    # Branch
    path("branch/", views.IPFabricBranchListView.as_view(), name="ipfabricbranch_list"),
    path(
        "branch/<int:pk>/", include(get_model_urls("ipfabric_netbox", "ipfabricbranch"))
    ),
    # Transform Map
    path(
        "transform-map/",
        views.IPFabricTransformMapListView.as_view(),
        name="ipfabrictransformmap_list",
    ),
    path(
        "transform-map/restore/",
        views.IPFabricTransformMapRestoreView.as_view(),
        name="ipfabrictransformmap_restore",
    ),
    path(
        "transform-map/add",
        views.IPFabricTransformMapEditView.as_view(),
        name="ipfabrictransformmap_add",
    ),
    path(
        "transform-map/delete/",
        views.IPFabricTransformMapBulkDeleteView.as_view(),
        name="ipfabrictransformmap_bulk_delete",
    ),
    path(
        "transform-map/<int:pk>/",
        include(get_model_urls("ipfabric_netbox", "ipfabrictransformmap")),
    ),
    path(
        "transform-map/<int:pk>/delete/",
        views.IPFabricTransformMapDeleteView.as_view(),
        name="ipfabrictransformmap_delete",
    ),
    # Transform field
    path(
        "transform-field/",
        views.IPFabricTransformFieldListView.as_view(),
        name="ipfabrictransformfield_list",
    ),
    path(
        "transform-field/add/",
        views.IPFabricTransformFieldEditView.as_view(),
        name="ipfabrictransformfield_add",
    ),
    path(
        "transform-field/<int:pk>/",
        include(get_model_urls("ipfabric_netbox", "ipfabrictransformfield")),
    ),
    path(
        "transform-field/<int:pk>/delete/",
        views.IPFabricTransformFieldDeleteView.as_view(),
        name="ipfabrictransformfield_delete",
    ),
    # Relationship Field
    path(
        "relationship-field/",
        views.IPFabricRelationshipFieldListView.as_view(),
        name="ipfabricrelationshipfield_list",
    ),
    path(
        "relationship-field/add/",
        views.IPFabricRelationshipFieldEditView.as_view(),
        name="ipfabricrelationshipfield_add",
    ),
    path(
        "relationship-field/<int:pk>/",
        include(get_model_urls("ipfabric_netbox", "ipfabricrelationshipfield")),
    ),
    path(
        "relationship-field/<int:pk>/delete/",
        views.IPFabricRelationshipFieldDeleteView.as_view(),
        name="ipfabricrelationshipfield_delete",
    ),
)



import logging

from django.core.cache import cache
from django.utils import timezone
from extras.choices import LogLevelChoices


class SyncLogging:
    def __init__(self, key_prefix="ipfabric_sync", job=None, cache_timeout=3600):
        self.key_prefix = key_prefix
        self.job_id = job
        self.cache_key = f"{self.key_prefix}_{job}"
        self.cache_timeout = cache_timeout
        self.log_data = {"logs": [], "statistics": {}}
        self.logger = logging.getLogger("ipfabric.sync")

    def _log(self, obj, message, level=LogLevelChoices.LOG_DEFAULT):
        """
        Log a message from a test method. Do not call this method directly; use one of the log_* wrappers below.
        """
        if level not in LogLevelChoices.values():
            raise Exception(f"Unknown logging level: {level}")
        self.log_data["logs"].append(
            (
                timezone.now().isoformat(),
                level,
                str(obj) if obj else None,
                obj.get_absolute_url() if hasattr(obj, "get_absolute_url") else None,
                message,
            )
        )
        cache.set(self.cache_key, self.log_data, self.cache_timeout)

    def log(self, message):
        """
        Log a message which is not associated with a particular object.
        """
        self._log(None, message, level=LogLevelChoices.LOG_DEFAULT)
        self.logger.info(message)

    def log_success(self, message, obj=None):
        """
        Record a successful test against an object.
        """
        self._log(obj, message, level=LogLevelChoices.LOG_SUCCESS)
        self.logger.info(f"Success | {obj}: {message}")

    def log_info(self, message: str, obj=None):
        """
        Log an informational message.
        """
        self._log(obj, message, level=LogLevelChoices.LOG_INFO)
        self.logger.info(f"Info | {obj}: {message}")

    def log_warning(self, message, obj=None):
        """
        Log a warning.
        """
        self._log(obj, message, level=LogLevelChoices.LOG_WARNING)
        self.logger.info(f"Warning | {obj}: {message}")

    def log_failure(self, message, obj=None):
        """
        Log a failure. Calling this method will automatically mark the report as failed.
        """
        self._log(obj, message, level=LogLevelChoices.LOG_FAILURE)
        self.logger.info(f"Failure | {obj}: {message}")

    def increment_statistics(self, model, current=None, total=None):
        statistics = self.log_data.get("statistics")

        if not statistics.get(model):
            stats = statistics[model] = {"current": current, "total": total}
        else:
            stats = statistics.get(model)
        if total:
            stats["total"] = total
        if current:
            stats["current"] = current
        cache.set(self.cache_key, self.log_data, self.cache_timeout)

    def clear_log(self):
        self.log_data["logs"] = []

    @classmethod
    def retrieve_from_cache(cls, key_prefix="log"):
        cache_key = f"{key_prefix}_log"
        log_data = cache.get(cache_key)
        if log_data is None:
            return cls(key_prefix)
        log = cls(key_prefix)
        log.log_data = log_data
        return log



from extras.plugins import PluginMenuButton
from extras.plugins import PluginMenuItem
from netbox.plugins import PluginMenu
from utilities.choices import ButtonColorChoices

sync_buttons = [
    PluginMenuButton(
        link="plugins:ipfabric_netbox:ipfabricsync_add",
        title="Add",
        icon_class="mdi mdi-plus-thick",
        permissions=["ipfabric_netbox.add_ipfabricsync"],
        color=ButtonColorChoices.GREEN,
    )
]

source_buttons = [
    PluginMenuButton(
        link="plugins:ipfabric_netbox:ipfabricsource_add",
        title="Add",
        icon_class="mdi mdi-plus-thick",
        permissions=["ipfabric_netbox.add_ipfabricsource"],
        color=ButtonColorChoices.GREEN,
    )
]

source = PluginMenuItem(
    link="plugins:ipfabric_netbox:ipfabricsource_list",
    link_text="Sources",
    buttons=source_buttons,
    permissions=["ipfabric_netbox.view_ipfabricsource"],
)

snapshot = PluginMenuItem(
    link="plugins:ipfabric_netbox:ipfabricsnapshot_list",
    link_text="Snapshots",
    permissions=["ipfabric_netbox.view_ipfabricsnapshot"],
)


ingestion = PluginMenuItem(
    link="plugins:ipfabric_netbox:ipfabricsync_list",
    link_text="Ingestion",
    buttons=sync_buttons,
    permissions=["ipfabric_netbox.view_ipfabricsync"],
)

tm = PluginMenuItem(
    link="plugins:ipfabric_netbox:ipfabrictransformmap_list",
    link_text="Transform Maps",
    permissions=["ipfabric_netbox.view_ipfabrictransformmap"],
    buttons=[
        PluginMenuButton(
            link="plugins:ipfabric_netbox:ipfabrictransformmap_add",
            title="Add",
            icon_class="mdi mdi-plus-thick",
            permissions=["ipfabric_netbox.add_ipfabrictransformmap"],
            color=ButtonColorChoices.GREEN,
        )
    ],
)
menu = PluginMenu(
    label="IP Fabric",
    icon_class="mdi mdi-cloud-sync",
    groups=(("IP Fabric", (source, snapshot, ingestion, tm)),),
)



import json
import uuid

from core.exceptions import SyncError
from dcim.models import Device
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.shortcuts import get_object_or_404
from django.utils.text import slugify
from django_tables2 import Column
from ipfabric import IPFClient
from jinja2.sandbox import SandboxedEnvironment
from netbox.config import get_config
from netutils.utils import jinja2_convenience_function

from ..choices import IPFabricSourceTypeChoices
from .nbutils import device_serial_max_length
from .nbutils import order_devices
from .nbutils import order_members
from .nbutils import order_pn
from .nbutils import order_vrf


def slugify_text(value):
    return slugify(value)


def serial(value):
    sn_length = len(value.get("sn"))
    serial_number = value.get("sn") if sn_length < device_serial_max_length else ""
    if not serial_number:
        serial_number = value.get("id")
    return serial_number


IPF_JINJA_FILTERS = {"slugify": slugify_text, "serial": serial}


def render_jinja2(template_code, context):
    """
    Render a Jinja2 template with the provided context. Return the rendered content.
    """
    environment = SandboxedEnvironment()
    environment.filters.update(get_config().JINJA2_FILTERS)
    environment.filters.update(IPF_JINJA_FILTERS)
    environment.filters.update(jinja2_convenience_function())
    return environment.from_string(source=template_code).render(**context)


class IPFabric(object):
    def __init__(self, parameters=None, transform_map=None) -> None:
        if parameters:
            self.ipf = IPFClient(**parameters, unloaded=True)
        else:
            self.ipf = IPFClient(
                **settings.PLUGINS_CONFIG["ipfabric_netbox"], unloaded=True
            )
        self.transform_map = transform_map

    def get_snapshots(self) -> dict:
        formatted_snapshots = {}
        if self.ipf:
            for snapshot_ref, snapshot in self.ipf.snapshots.items():
                if snapshot.status != "done" and snapshot.finish_status != "done":
                    continue
                if snapshot_ref in ["$prev", "$lastLocked"]:
                    continue
                if snapshot.name:
                    description = (
                        snapshot.name
                        + " - "
                        + snapshot.end.strftime("%d-%b-%y %H:%M:%S")
                    )
                else:
                    description = snapshot.end.strftime("%d-%b-%y %H:%M:%S")

                formatted_snapshots[snapshot_ref] = (description, snapshot.snapshot_id)
        return formatted_snapshots

    def get_sites(self, snapshot=None) -> dict():
        if snapshot:
            raw_sites = self.ipf.inventory.sites.all(snapshot_id=snapshot)
        else:
            raw_sites = self.ipf.inventory.sites.all()
        sites = []
        for item in raw_sites:
            sites.append(item["siteName"])
        return sites

    def get_table_data(self, table, device):
        filter = {"sn": ["eq", device.serial]}
        split = table.split(".")

        if len(split) == 2:
            if split[1] == "serial_ports":
                table = getattr(self.ipf.technology, split[1])
            else:
                tech = getattr(self.ipf.technology, split[0])
                table = getattr(tech, split[1])
        else:
            table = getattr(self.ipf.inventory, split[0])

        columns = self.ipf.get_columns(table.endpoint)

        columns.pop(0)

        columns = [(k, Column()) for k in columns]
        data = table.all(
            filters=filter,
        )
        return data, columns


class IPFabricSyncRunner(object):
    def __init__(
        self, transform_map, sync=None, client: IPFabric = None, settings: dict = None
    ) -> None:
        self.client = client
        self.settings = settings
        self.transform_map = transform_map
        self.sync = sync
        self.relationship_store = {}
        self.siteUUID = {}
        if hasattr(self.sync, "logger"):
            self.logger = self.sync.logger
        self.interface_count_total = 0
        self.interface_count = 1

        if self.sync.snapshot_data.status != "loaded":
            raise SyncError("Snapshot not loaded in IP Fabric.")

    def get_model_or_update(self, app, model, data, uuid=None):
        transform_map = self.transform_map.objects.filter(
            target_model__app_label=app, target_model__model=model
        ).first()

        if not transform_map:
            raise SystemError(f"No transform map available for {app}: {model}")

        model_settings = self.settings.get(model, False)
        object = None

        if model_settings:
            print(f"Creating {model}")
            object = transform_map.update_or_create_instance(
                data=data,
                uuid=uuid,
                relationship_store=self.relationship_store,
                tags=self.sync.tags.all(),
                logger=self.logger,
            )
        else:
            print(f"Getting {model}")
            coalesce_fields = transform_map.get_coalesce_fields(data)
            object = get_object_or_404(
                transform_map.target_model.model_class().objects.all(),
                **coalesce_fields,
            )

        store = self.relationship_store.get(uuid)

        if store:
            store[object._meta.model] = object
        else:
            self.relationship_store[uuid] = {object._meta.model: object}

        return object

    def create_interface(
        self, device_interface, device_uuid, managed_ips, device_object, device
    ):
        device_interface["loginIp"] = device.get("loginIp")
        interface_object = self.get_model_or_update(
            "dcim", "interface", device_interface, uuid=device_uuid
        )

        self.logger.increment_statistics(
            model="interface",
            current=self.interface_count,
            total=self.interface_count_total,
        )
        self.interface_count += 1

        if self.settings.get("ipaddress"):
            managed_ip = managed_ips.get(device_object.serial, {}).get(
                interface_object.name
            )
            if managed_ip:
                ip_address_obj = self.get_model_or_update(
                    "ipam",
                    "ipaddress",
                    managed_ip,
                )
                try:
                    other_device = Device.objects.get(primary_ip4=ip_address_obj)
                    if other_device and device_object != other_device:
                        other_device.primary_ip4 = None
                        other_device.save()
                except ObjectDoesNotExist:
                    pass

                if device.get("loginIp") == device_interface.get("primaryIp"):
                    device_object.primary_ip4 = ip_address_obj
                    device_object.save()

        return True

    def collect_data(self):
        try:
            self.logger.log_info(
                "Collecting information from IP Fabric",
                obj=self.sync.snapshot_data.source,
            )
            data = {}
            if self.sync.snapshot_data.source.type == IPFabricSourceTypeChoices.REMOTE:
                self.logger.log_info(
                    "Remote collector checking for snapshot data.", obj=self.sync
                )
                if not self.sync.snapshot_data.ipf_data.count() > 0:
                    raise SyncError(
                        "No snapshot data available. This is a remote sync. Push data to NetBox first."
                    )
                data["site"] = list(
                    self.sync.snapshot_data.ipf_data.filter(type="site").values_list(
                        "data", flat=True
                    )
                )
                data["device"] = list(
                    self.sync.snapshot_data.ipf_data.filter(type="device").values_list(
                        "data", flat=True
                    )
                )
                data["virtualchassis"] = list(
                    self.sync.snapshot_data.ipf_data.filter(
                        type="virtualchassis"
                    ).values_list("data", flat=True)
                )
                data["interface"] = list(
                    self.sync.snapshot_data.ipf_data.filter(
                        type="interface"
                    ).values_list("data", flat=True)
                )
                data["inventoryitem"] = list(
                    self.sync.snapshot_data.ipf_data.filter(
                        type="inventoryitem"
                    ).values_list("data", flat=True)
                )
                data["vlan"] = list(
                    self.sync.snapshot_data.ipf_data.filter(type="vlan").values_list(
                        "data", flat=True
                    )
                )
                data["vrf"] = list(
                    self.sync.snapshot_data.ipf_data.filter(type="vrf").values_list(
                        "data", flat=True
                    )
                )
                data["prefix"] = list(
                    self.sync.snapshot_data.ipf_data.filter(type="prefix").values_list(
                        "data", flat=True
                    )
                )
                data["ipaddress"] = list(
                    self.sync.snapshot_data.ipf_data.filter(
                        type="ipaddress"
                    ).values_list("data", flat=True)
                )
            else:
                self.logger.log_info(
                    "Local collector being used for snapshot data.", obj=self.sync
                )
                excluded_vendors = ["aws", "azure"]

                filter = {
                    "and": [{"vendor": ["neq", vendor]} for vendor in excluded_vendors]
                }
                # filter = {"and": [{"vendor": ["neq", "aws"]}, {"vendor": ["neq", "azure"]}]}

                if ingestion_sites := self.settings.get("sites"):
                    site_filter = {
                        "or": [{"siteName": ["eq", site]} for site in ingestion_sites]
                    }
                    filter["and"].append(site_filter)

                    self.logger.log_info(
                        f"Creating site filter {json.dumps(site_filter)}", obj=self.sync
                    )
                else:
                    site_filter = {}

                data["site"] = self.client.inventory.sites.all(
                    snapshot_id=self.settings["snapshot_id"]
                )

                data["device"] = self.client.inventory.devices.all(
                    snapshot_id=self.settings["snapshot_id"], filters=filter
                )

                data["virtualchassis"] = (
                    self.client.technology.platforms.stacks_members.all(
                        snapshot_id=self.settings["snapshot_id"], filters=site_filter
                    )
                )

                data["interface"] = self.client.inventory.interfaces.all(
                    snapshot_id=self.settings["snapshot_id"]
                )

                data["inventoryitem"] = self.client.inventory.pn.all(
                    snapshot_id=self.settings["snapshot_id"],
                    filters={
                        "and": [{"sn": ["empty", False]}, {"name": ["empty", False]}]
                    },
                )

                data["vlan"] = self.client.technology.vlans.site_summary.all(
                    snapshot_id=self.settings["snapshot_id"], filters=site_filter
                )

                data["vrf"] = self.client.technology.routing.vrf_detail.all(
                    snapshot_id=self.settings["snapshot_id"], filters=site_filter
                )

                if site_filter:
                    networks_filter = {
                        "and": [site_filter, {"and": [{"net": ["empty", False]}]}]
                    }
                else:
                    networks_filter = {"and": [{"net": ["empty", False]}]}
                self.logger.log_info(f"Creating network filter: `{networks_filter}`")
                data["prefix"] = self.client.technology.managed_networks.networks.all(
                    snapshot_id=self.settings["snapshot_id"], filters=networks_filter
                )

                data["ipaddress"] = (
                    self.client.technology.addressing.managed_ip_ipv4.all(
                        snapshot_id=self.settings["snapshot_id"]
                    )
                )
        except Exception as e:
            self.logger.log_failure(
                f"Error collecting data from IP Fabric: {e}", obj=self.sync
            )
            raise SyncError(f"Error collecting data from IP Fabric: {e}")

        self.logger.log_info(
            f"{len(data['site'])} sites collected", obj=self.sync.snapshot_data.source
        )
        self.logger.log_info(
            f"{len(data['device'])} devices collected",
            obj=self.sync.snapshot_data.source,
        )
        self.logger.log_info(
            f"{len(data['virtualchassis'])} stack members collected",
            obj=self.sync.snapshot_data.source,
        )

        self.logger.log_info(
            f"{len(data['interface'])} interfaces collected",
            obj=self.sync.snapshot_data.source,
        )

        self.logger.log_info(
            f"{len(data.get('inventoryitem', []))} part numbers collected",
            obj=self.sync.snapshot_data.source,
        )

        self.logger.log_info(
            f"{len(data.get('vlan', []))} VLANs collected",
            obj=self.sync.snapshot_data.source,
        )

        self.logger.log_info(
            f"{len(data.get('vrf', []))} VRFs collected",
            obj=self.sync.snapshot_data.source,
        )

        self.logger.log_info(
            f"{len(data.get('prefix', []))} networks collected",
            obj=self.sync.snapshot_data.source,
        )

        self.logger.log_info(
            f"{len(data.get('ipaddress', []))} management IP's collected",
            obj=self.sync.snapshot_data.source,
        )

        self.logger.log_info("Ordering devices", obj=self.sync)

        members = order_members(data.get("virtualchassis", []))
        devices = order_devices(data.get("device", []), members)

        self.logger.log_info("Ordering Part Numbers", obj=self.sync)

        part_numbers = order_pn(data.get("inventoryitem", []))

        self.logger.log_info("Ordering VRF's", obj=self.sync)

        vrfs = order_vrf(data["vrf"])

        managed_ips = {}
        site_dict = {}
        interface_dict = {}
        for site in data["site"]:
            site_dict[site["siteName"]] = site

        for interface in data["interface"]:
            if int_sn := interface.get("sn"):
                if interface_dict.get(int_sn):
                    interface_dict[int_sn].append(interface)
                else:
                    interface_dict[int_sn] = [interface]

        for ip in data["ipaddress"]:
            if managed_ips.get(ip["sn"]):
                managed_ips[ip["sn"]][ip["intName"]] = ip
            else:
                managed_ips[ip["sn"]] = {ip["intName"]: ip}

        return (
            site_dict,
            devices,
            interface_dict,
            part_numbers,
            vrfs,
            data["vlan"],
            data["prefix"],
            managed_ips,
        )

    def sync_devices(self, branch=None):
        self.logger.log_info("Starting device sync", obj=self.sync)

        (
            site_dict,
            devices,
            interface_dict,
            part_numbers,
            vrfs,
            vlans,
            networks,
            managed_ips,
        ) = self.collect_data()
        vlan_count = 1
        inventoryitem_count = 1
        vrf_count = 1
        network_count = 1
        device_vrfs_total = 0

        for device_count, device in enumerate(devices, start=1):
            print(device)
            print(f"Device {device_count} out of {len(devices)}")
            self.logger.increment_statistics(
                model="device", current=device_count, total=len(devices)
            )

            device_uuid = str(uuid.uuid4())

            site_object = self.get_model_or_update(
                "dcim", "site", site_dict[device["siteName"]], uuid=device_uuid
            )

            self.get_model_or_update("dcim", "manufacturer", device, uuid=device_uuid)
            self.get_model_or_update("dcim", "devicetype", device, uuid=device_uuid)

            self.get_model_or_update("dcim", "platform", device, uuid=device_uuid)

            self.get_model_or_update("dcim", "devicerole", device, uuid=device_uuid)

            device_object = self.get_model_or_update(
                "dcim", "device", device, uuid=device_uuid
            )

            site_object.custom_field_data["ipfabric_source"] = (
                self.sync.snapshot_data.source.pk
            )

            device_object.custom_field_data["ipfabric_source"] = (
                self.sync.snapshot_data.source.pk
            )
            if branch:
                site_object.custom_field_data["ipfabric_branch"] = branch.pk
                device_object.custom_field_data["ipfabric_branch"] = branch.pk

            site_object.save()
            device_object.save()

            if self.settings.get("virtualchassis"):
                if member := device.get("virtual_chassis"):
                    self.get_model_or_update("dcim", "virtualchassis", member)
                    device_object = self.get_model_or_update(
                        "dcim", "device", device, uuid=device_uuid
                    )

            if device_object and self.settings.get("interface"):
                device_interfaces = interface_dict.get(device.get("sn"), [])
                self.interface_count_total += len(device_interfaces)
                for device_interface in device_interfaces:
                    self.create_interface(
                        device_interface,
                        device_uuid,
                        managed_ips,
                        device_object,
                        device,
                    )
                    # x = threading.Thread(target=self.create_interface, args=((device_interface, device_uuid, managed_ips, device_object, device)))
                    # threads.append(x)
                    # x.start()

            if device_object and self.settings.get("inventoryitem"):
                device_parts = part_numbers.get(device_object.serial, [])
                for part in device_parts:
                    self.get_model_or_update(
                        "dcim", "inventoryitem", part, uuid=device_uuid
                    )
                    self.logger.increment_statistics(
                        model="inventory_item",
                        current=inventoryitem_count,
                        total=len(part_numbers),
                    )
                    inventoryitem_count += 1

            if self.settings.get("vrf"):
                device_vrfs = vrfs.get(device_object.serial, [])
                device_vrfs_total += len(device_vrfs)
                for vrf in device_vrfs:
                    self.get_model_or_update("ipam", "vrf", vrf, uuid=device_uuid)
                    self.logger.increment_statistics(
                        model="vrf", current=vrf_count, total=device_vrfs_total
                    )
                    vrf_count += 1

            device_count += 1

        if self.settings.get("vlan"):
            for vlan in vlans:
                self.get_model_or_update("ipam", "vlan", vlan)
                self.logger.increment_statistics(
                    model="vlan", current=vlan_count, total=len(vlans)
                )
                vlan_count += 1

        if self.settings.get("prefix"):
            for network in networks:
                self.get_model_or_update("ipam", "prefix", network)
                self.logger.increment_statistics(
                    model="prefix", current=network_count, total=len(networks)
                )
                network_count += 1

    def sync(self):
        self.sync_devices()



import copy
import json
import uuid

from dcim.models import Device
from django.contrib.contenttypes.models import ContentType
from django.core import serializers
from django.test import TestCase
from django.utils import timezone

from ipfabric_netbox.models import IPFabricSnapshot
from ipfabric_netbox.models import IPFabricSource
from ipfabric_netbox.models import IPFabricSync
from ipfabric_netbox.models import IPFabricTransformField
from ipfabric_netbox.models import IPFabricTransformMap
from ipfabric_netbox.utilities.ipfutils import IPFabricSyncRunner

transform_maps = [
    {
        "model": "ipfabric_netbox.ipfabrictransformmap",
        "pk": 4,
        "fields": {
            "created": "2023-06-22T09:12:03.018Z",
            "last_updated": "2023-06-22T09:12:03.018Z",
            "custom_field_data": {},
            "name": "Platform Transform Map",
            "source_model": "device",
            "target_model": 43,
            "status": "",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformmap",
        "pk": 1,
        "fields": {
            "created": "2023-06-22T09:12:02.886Z",
            "last_updated": "2023-06-22T09:53:32.100Z",
            "custom_field_data": {},
            "name": "Site Transform Map",
            "source_model": "site",
            "target_model": 56,
            "status": "active",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformmap",
        "pk": 2,
        "fields": {
            "created": "2023-06-22T09:12:02.924Z",
            "last_updated": "2023-06-22T15:02:20.038Z",
            "custom_field_data": {},
            "name": "Manufacturer Transform Map",
            "source_model": "device",
            "target_model": 42,
            "status": "",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformmap",
        "pk": 5,
        "fields": {
            "created": "2023-06-22T09:12:03.087Z",
            "last_updated": "2023-06-22T15:51:28.987Z",
            "custom_field_data": {},
            "name": "Device Transform Map",
            "source_model": "device",
            "target_model": 31,
            "status": "active",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformmap",
        "pk": 6,
        "fields": {
            "created": "2023-06-22T09:12:03.134Z",
            "last_updated": "2023-06-22T16:24:22.879Z",
            "custom_field_data": {},
            "name": "Device Role Transform Map",
            "source_model": "device",
            "target_model": 34,
            "status": "",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformmap",
        "pk": 3,
        "fields": {
            "created": "2023-06-22T09:12:02.971Z",
            "last_updated": "2023-06-23T13:23:50.445Z",
            "custom_field_data": {},
            "name": "Device Type Transform Map",
            "source_model": "device",
            "target_model": 35,
            "status": "",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformmap",
        "pk": 10,
        "fields": {
            "created": "2023-06-23T14:42:40.118Z",
            "last_updated": "2023-06-23T14:42:40.118Z",
            "custom_field_data": {},
            "name": "Interface Transform Map",
            "source_model": "interface",
            "target_model": 38,
            "status": "",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformmap",
        "pk": 11,
        "fields": {
            "created": "2023-06-28T15:13:56.033Z",
            "last_updated": "2023-06-28T15:18:37.423Z",
            "custom_field_data": {},
            "name": "Inventory Transform Map",
            "source_model": "part_number",
            "target_model": 40,
            "status": "",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformmap",
        "pk": 12,
        "fields": {
            "created": "2023-07-14T15:37:25.428Z",
            "last_updated": "2023-07-14T15:37:25.428Z",
            "custom_field_data": {},
            "name": "VLAN Transform Map",
            "source_model": "vlan",
            "target_model": 75,
            "status": "active",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformmap",
        "pk": 13,
        "fields": {
            "created": "2023-07-14T16:39:14.674Z",
            "last_updated": "2023-07-14T16:39:14.675Z",
            "custom_field_data": {},
            "name": "VRF Transform Map",
            "source_model": "vrf",
            "target_model": 73,
            "status": "",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformmap",
        "pk": 14,
        "fields": {
            "created": "2023-07-18T09:36:57.588Z",
            "last_updated": "2023-07-18T09:36:57.588Z",
            "custom_field_data": {},
            "name": "Prefix Transform Map",
            "source_model": "prefix",
            "target_model": 69,
            "status": "active",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformmap",
        "pk": 15,
        "fields": {
            "created": "2023-07-31T10:54:29.792Z",
            "last_updated": "2023-07-31T10:54:29.792Z",
            "custom_field_data": {},
            "name": "Virtual Chassis Transform Map",
            "source_model": "virtualchassis",
            "target_model": 58,
            "status": "",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformmap",
        "pk": 16,
        "fields": {
            "created": "2023-08-15T14:34:00.334Z",
            "last_updated": "2023-08-15T14:34:00.334Z",
            "custom_field_data": {},
            "name": "IP Address Transform Map",
            "source_model": "ipaddress",
            "target_model": 68,
            "status": "active",
        },
    },
]

transform_fields = [
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 19,
        "fields": {
            "created": "2023-06-23T15:01:13.459Z",
            "last_updated": "2023-06-23T15:01:13.459Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 10,
            "source_field": "dscr",
            "target_field": "description",
            "coalesce": False,
            "template": '{{ object.dscr or ""}}',
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 20,
        "fields": {
            "created": "2023-06-28T10:39:24.588Z",
            "last_updated": "2023-06-28T10:39:24.588Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 10,
            "source_field": "media",
            "target_field": "type",
            "coalesce": False,
            "template": "1000base-t",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 2,
        "fields": {
            "created": "2023-06-22T09:12:02.900Z",
            "last_updated": "2023-06-22T09:12:02.900Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 1,
            "source_field": "siteName",
            "target_field": "slug",
            "coalesce": True,
            "template": "{{ object.siteName | slugify }}",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 3,
        "fields": {
            "created": "2023-06-22T09:12:02.933Z",
            "last_updated": "2023-06-22T09:12:02.933Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 2,
            "source_field": "vendor",
            "target_field": "name",
            "coalesce": False,
            "template": None,
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 25,
        "fields": {
            "created": "2023-06-28T10:46:04.153Z",
            "last_updated": "2023-06-28T10:46:04.153Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 10,
            "source_field": "l1",
            "target_field": "enabled",
            "coalesce": False,
            "template": '{% if object.l1 == "up" %}True{% else %}False{% endif %}',
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 34,
        "fields": {
            "created": "2023-07-14T15:38:30.728Z",
            "last_updated": "2023-07-31T10:37:21.182Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 12,
            "source_field": "vlanName",
            "target_field": "name",
            "coalesce": False,
            "template": '{{ object.vlanName or ""}}',
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 40,
        "fields": {
            "created": "2023-07-31T15:07:09.911Z",
            "last_updated": "2023-07-31T15:07:09.911Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 15,
            "source_field": "master",
            "target_field": "name",
            "coalesce": True,
            "template": "",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 22,
        "fields": {
            "created": "2023-06-28T10:41:47.345Z",
            "last_updated": "2023-06-28T12:08:47.704Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 10,
            "source_field": "mac",
            "target_field": "mac_address",
            "coalesce": False,
            "template": '{% if object.mac %}{{ object.mac | mac_to_format(frmt="MAC_COLON_TWO") | upper }}{% else %}{{ "00:00:00:00:00:01" | mac_to_format(frmt="MAC_COLON_TWO") | upper }}{% endif %}',
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 4,
        "fields": {
            "created": "2023-06-22T09:12:02.948Z",
            "last_updated": "2023-06-22T09:12:02.948Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 2,
            "source_field": "vendor",
            "target_field": "slug",
            "coalesce": True,
            "template": "{{ object.vendor | slugify }}",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 8,
        "fields": {
            "created": "2023-06-22T09:12:03.043Z",
            "last_updated": "2023-07-14T10:54:57.940Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 4,
            "source_field": "vendor",
            "target_field": "slug",
            "coalesce": True,
            "template": "{{ object.vendor | slugify }}{%if object.family %}_{{ object.family | slugify }}{% endif %}",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 6,
        "fields": {
            "created": "2023-06-22T09:12:02.995Z",
            "last_updated": "2023-06-22T09:12:02.995Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 3,
            "source_field": "vendor",
            "target_field": "slug",
            "coalesce": True,
            "template": "{{ object.vendor | slugify }}",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 26,
        "fields": {
            "created": "2023-06-28T10:49:23.845Z",
            "last_updated": "2023-06-28T14:50:04.953Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 10,
            "source_field": "intName",
            "target_field": "name",
            "coalesce": True,
            "template": "",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 28,
        "fields": {
            "created": "2023-06-28T15:20:02.726Z",
            "last_updated": "2023-06-28T15:20:08.343Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 11,
            "source_field": "sn",
            "target_field": "serial",
            "coalesce": True,
            "template": "",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 29,
        "fields": {
            "created": "2023-06-28T15:20:22.262Z",
            "last_updated": "2023-06-28T15:20:22.262Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 11,
            "source_field": "pid",
            "target_field": "part_id",
            "coalesce": False,
            "template": "",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 7,
        "fields": {
            "created": "2023-06-22T09:12:03.027Z",
            "last_updated": "2023-07-14T10:56:34.746Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 4,
            "source_field": "family",
            "target_field": "name",
            "coalesce": False,
            "template": "{%if object.family %}{{ object.family | slugify }}{% else %}{{ object.vendor }}{% endif %}",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 35,
        "fields": {
            "created": "2023-07-14T15:38:43.642Z",
            "last_updated": "2023-07-14T15:39:37.833Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 12,
            "source_field": "vlanId",
            "target_field": "vid",
            "coalesce": True,
            "template": "",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 36,
        "fields": {
            "created": "2023-07-14T15:39:00.515Z",
            "last_updated": "2023-07-14T15:57:48.819Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 12,
            "source_field": "dscr",
            "target_field": "description",
            "coalesce": False,
            "template": '{{ object.dscr or ""}}',
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 11,
        "fields": {
            "created": "2023-06-22T09:12:03.112Z",
            "last_updated": "2023-06-22T09:12:03.112Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 5,
            "source_field": "sn",
            "target_field": "serial",
            "coalesce": True,
            "template": "{{ object | serial }}",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 37,
        "fields": {
            "created": "2023-07-14T16:39:34.015Z",
            "last_updated": "2023-07-14T16:39:34.015Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 13,
            "source_field": "vrf",
            "target_field": "name",
            "coalesce": True,
            "template": "",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 21,
        "fields": {
            "created": "2023-06-28T10:41:03.607Z",
            "last_updated": "2023-08-08T14:44:52.118Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 10,
            "source_field": "mtu",
            "target_field": "mtu",
            "coalesce": False,
            "template": "{{ object.mtu or 1500 }}",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 38,
        "fields": {
            "created": "2023-07-14T16:39:57.563Z",
            "last_updated": "2023-07-14T16:44:39.864Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 13,
            "source_field": "rd",
            "target_field": "rd",
            "coalesce": True,
            "template": "",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 42,
        "fields": {
            "created": "2023-08-09T12:31:17.235Z",
            "last_updated": "2023-08-09T12:31:17.235Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 5,
            "source_field": "hostname",
            "target_field": "vc_position",
            "coalesce": False,
            "template": "{% if object.virtual_chassis is defined %}{{ object.virtual_chassis.member }}{% else %}None{% endif %}",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 10,
        "fields": {
            "created": "2023-06-22T09:12:03.095Z",
            "last_updated": "2023-08-14T09:56:03.044Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 5,
            "source_field": "hostname",
            "target_field": "name",
            "coalesce": False,
            "template": "",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 1,
        "fields": {
            "created": "2023-06-22T09:12:02.889Z",
            "last_updated": "2023-07-17T10:08:24.194Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 1,
            "source_field": "siteName",
            "target_field": "name",
            "coalesce": False,
            "template": "",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 27,
        "fields": {
            "created": "2023-06-28T15:19:45.005Z",
            "last_updated": "2023-06-28T16:18:37.860Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 11,
            "source_field": "name",
            "target_field": "name",
            "coalesce": False,
            "template": "{% if object.name is not none %}{{ object.name }}{% elif object.dscr is not none %}{{ object.dscr}}{% else %}Default Name{% endif %}",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 12,
        "fields": {
            "created": "2023-06-22T09:12:03.144Z",
            "last_updated": "2023-06-22T16:24:30.681Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 6,
            "source_field": "devType",
            "target_field": "name",
            "coalesce": False,
            "template": "",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 13,
        "fields": {
            "created": "2023-06-22T09:12:03.161Z",
            "last_updated": "2023-06-22T16:24:34.880Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 6,
            "source_field": "devType",
            "target_field": "slug",
            "coalesce": True,
            "template": "{{ object.devType | slugify }}",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 39,
        "fields": {
            "created": "2023-07-18T09:40:37.112Z",
            "last_updated": "2023-07-18T09:41:09.238Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 14,
            "source_field": "net",
            "target_field": "prefix",
            "coalesce": True,
            "template": "",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 43,
        "fields": {
            "created": "2023-08-15T14:25:31.706Z",
            "last_updated": "2023-08-15T14:25:31.706Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 10,
            "source_field": "primaryIp",
            "target_field": "mgmt_only",
            "coalesce": False,
            "template": "{% if object.primaryIp == object.loginIp %}True{% else %}False{% endif %}",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 5,
        "fields": {
            "created": "2023-06-22T09:12:02.979Z",
            "last_updated": "2023-07-14T13:51:37.370Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 3,
            "source_field": "model",
            "target_field": "model",
            "coalesce": False,
            "template": '{% if object.model|string != ""%}{{ object.model | string }}{% else %}{{ object.vendor }} - {{ object.family }} - {{ object.platform }}{% endif %}',
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 14,
        "fields": {
            "created": "2023-06-22T09:12:03.177Z",
            "last_updated": "2023-06-22T16:24:48.259Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 6,
            "source_field": "devType",
            "target_field": "vm_role",
            "coalesce": True,
            "template": "False",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 44,
        "fields": {
            "created": "2023-08-15T14:38:13.622Z",
            "last_updated": "2023-08-15T20:12:45.341Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 16,
            "source_field": "sn",
            "target_field": "assigned_object_id",
            "coalesce": False,
            "template": "{{ dcim.Interface.objects.filter(device__serial=object.sn, name=object.intName).first().pk }}",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabrictransformfield",
        "pk": 45,
        "fields": {
            "created": "2023-08-15T14:59:24.131Z",
            "last_updated": "2023-08-15T20:26:08.071Z",
            "custom_field_data": {},
            "description": "",
            "comments": "",
            "transform_map": 16,
            "source_field": "net",
            "target_field": "address",
            "coalesce": True,
            "template": "{% if object.net %}{% set MASK = object.net.split('/')[1] %}{% else %}{% set MASK = 32 %}{% endif %}{{ object.ip }}/{{MASK}}",
        },
    },
]

transform_relationship_fields = [
    {
        "model": "ipfabric_netbox.ipfabricrelationshipfield",
        "pk": 21,
        "fields": {
            "created": "2023-07-31T19:48:34.485Z",
            "last_updated": "2023-08-09T10:41:49.273Z",
            "custom_field_data": {},
            "transform_map": 15,
            "source_model": 31,
            "target_field": "master",
            "coalesce": False,
            "template": "{{ dcim.Device.objects.filter(serial=object.sn).first().pk }}",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabricrelationshipfield",
        "pk": 3,
        "fields": {
            "created": "2023-06-23T11:01:56.160Z",
            "last_updated": "2023-06-23T11:04:55.366Z",
            "custom_field_data": {},
            "transform_map": 5,
            "source_model": 43,
            "target_field": "platform",
            "coalesce": False,
            "template": None,
        },
    },
    {
        "model": "ipfabric_netbox.ipfabricrelationshipfield",
        "pk": 4,
        "fields": {
            "created": "2023-06-23T11:05:37.165Z",
            "last_updated": "2023-06-23T11:05:37.165Z",
            "custom_field_data": {},
            "transform_map": 5,
            "source_model": 56,
            "target_field": "site",
            "coalesce": False,
            "template": None,
        },
    },
    {
        "model": "ipfabric_netbox.ipfabricrelationshipfield",
        "pk": 5,
        "fields": {
            "created": "2023-06-23T11:07:26.908Z",
            "last_updated": "2023-06-23T11:07:26.908Z",
            "custom_field_data": {},
            "transform_map": 5,
            "source_model": 35,
            "target_field": "device_type",
            "coalesce": False,
            "template": None,
        },
    },
    {
        "model": "ipfabric_netbox.ipfabricrelationshipfield",
        "pk": 6,
        "fields": {
            "created": "2023-06-23T11:11:42.914Z",
            "last_updated": "2023-06-23T11:11:42.914Z",
            "custom_field_data": {},
            "transform_map": 5,
            "source_model": 34,
            "target_field": "device_role",
            "coalesce": False,
            "template": None,
        },
    },
    {
        "model": "ipfabric_netbox.ipfabricrelationshipfield",
        "pk": 7,
        "fields": {
            "created": "2023-06-23T13:42:05.285Z",
            "last_updated": "2023-06-23T13:42:05.285Z",
            "custom_field_data": {},
            "transform_map": 3,
            "source_model": 42,
            "target_field": "manufacturer",
            "coalesce": False,
            "template": None,
        },
    },
    {
        "model": "ipfabric_netbox.ipfabricrelationshipfield",
        "pk": 9,
        "fields": {
            "created": "2023-06-23T14:17:57.664Z",
            "last_updated": "2023-06-23T14:17:57.664Z",
            "custom_field_data": {},
            "transform_map": 4,
            "source_model": 42,
            "target_field": "manufacturer",
            "coalesce": False,
            "template": None,
        },
    },
    {
        "model": "ipfabric_netbox.ipfabricrelationshipfield",
        "pk": 22,
        "fields": {
            "created": "2023-08-09T12:39:35.028Z",
            "last_updated": "2023-08-09T12:39:35.028Z",
            "custom_field_data": {},
            "transform_map": 5,
            "source_model": 58,
            "target_field": "virtual_chassis",
            "coalesce": False,
            "template": "{% if object.virtual_chassis is defined %}{{ dcim.VirtualChassis.objects.filter(name=object.virtual_chassis.master).first().pk }}{% endif %}",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabricrelationshipfield",
        "pk": 11,
        "fields": {
            "created": "2023-06-28T11:00:44.494Z",
            "last_updated": "2023-06-28T14:37:54.113Z",
            "custom_field_data": {},
            "transform_map": 10,
            "source_model": 31,
            "target_field": "device",
            "coalesce": True,
            "template": None,
        },
    },
    {
        "model": "ipfabric_netbox.ipfabricrelationshipfield",
        "pk": 12,
        "fields": {
            "created": "2023-06-28T15:19:18.237Z",
            "last_updated": "2023-06-28T15:19:18.237Z",
            "custom_field_data": {},
            "transform_map": 11,
            "source_model": 42,
            "target_field": "manufacturer",
            "coalesce": False,
            "template": None,
        },
    },
    {
        "model": "ipfabric_netbox.ipfabricrelationshipfield",
        "pk": 13,
        "fields": {
            "created": "2023-06-28T15:19:35.397Z",
            "last_updated": "2023-07-14T14:15:42.974Z",
            "custom_field_data": {},
            "transform_map": 11,
            "source_model": 31,
            "target_field": "device",
            "coalesce": False,
            "template": None,
        },
    },
    {
        "model": "ipfabric_netbox.ipfabricrelationshipfield",
        "pk": 14,
        "fields": {
            "created": "2023-07-14T15:39:27.858Z",
            "last_updated": "2023-07-14T15:39:27.858Z",
            "custom_field_data": {},
            "transform_map": 12,
            "source_model": 56,
            "target_field": "site",
            "coalesce": True,
            "template": None,
        },
    },
    {
        "model": "ipfabric_netbox.ipfabricrelationshipfield",
        "pk": 23,
        "fields": {
            "created": "2023-08-15T14:48:42.409Z",
            "last_updated": "2023-08-15T20:24:33.161Z",
            "custom_field_data": {},
            "transform_map": 16,
            "source_model": 5,
            "target_field": "assigned_object_type",
            "coalesce": False,
            "template": '{{ contenttypes.ContentType.objects.get(app_label="dcim", model="interface").pk }}',
        },
    },
    {
        "model": "ipfabric_netbox.ipfabricrelationshipfield",
        "pk": 24,
        "fields": {
            "created": "2023-08-15T21:44:41.058Z",
            "last_updated": "2023-08-15T21:47:26.684Z",
            "custom_field_data": {},
            "transform_map": 16,
            "source_model": 73,
            "target_field": "vrf",
            "coalesce": False,
            "template": "{{ ipam.VRF.objects.filter(name=object.vrf).first().pk }}",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabricrelationshipfield",
        "pk": 15,
        "fields": {
            "created": "2023-07-18T09:40:50.275Z",
            "last_updated": "2023-08-15T21:53:36.179Z",
            "custom_field_data": {},
            "transform_map": 14,
            "source_model": 56,
            "target_field": "site",
            "coalesce": True,
            "template": "{% set SLUG = object.siteName | slugify %}\r\n{{ dcim.Site.objects.filter(slug=SLUG).first().pk }}",
        },
    },
    {
        "model": "ipfabric_netbox.ipfabricrelationshipfield",
        "pk": 17,
        "fields": {
            "created": "2023-07-18T21:19:36.757Z",
            "last_updated": "2023-07-18T21:19:44.148Z",
            "custom_field_data": {},
            "transform_map": 14,
            "source_model": 73,
            "target_field": "vrf",
            "coalesce": True,
            "template": "{{ ipam.VRF.objects.filter(name=object.vrf).first().pk }}",
        },
    },
]


class IPFabricTransformMapModelTestCase(TestCase):
    def setUp(self):
        deserialized_tm = serializers.deserialize("json", json.dumps(transform_maps))
        for deserialized_object in deserialized_tm:
            deserialized_object.save()

        deserialized_tf = serializers.deserialize("json", json.dumps(transform_fields))
        for deserialized_object in deserialized_tf:
            deserialized_object.save()

        deserialized_trf = serializers.deserialize(
            "json", json.dumps(transform_relationship_fields)
        )
        for deserialized_object in deserialized_trf:
            deserialized_object.save()

        source = IPFabricSource.objects.create(
            name="test",
            url="https://localhost",
            status="new",
            parameters={"auth": "token123", "verify": True},
        )
        snapshot = IPFabricSnapshot.objects.create(
            name="S01 - Day 2 - IPF Lab - 02-Jul-21 06:29:16 - 12dd8c61-129c-431a-b98b-4c9211571f89",
            source=source,
            snapshot_id="12dd8c61-129c-431a-b98b-4c9211571f89",
            data={
                "end": "2021-07-02T06:29:16.311000Z",
                "name": "S01 - Day 2 - IPF Lab",
                "note": "Multi-Environment containing:\\nAWS, Azure, ACI, NSX-T, Viptela, Versa, SilverPeak, Meraki\\n\\nSite 48 - devices added, NTP issue\\nSite 68 &38 - NTP update\\nSite 38 - resiliency affected (no ospfx2 - no L1 link x1) + passive interfaces FIXED / NTP partial update\\n?E2E: 38 - 66 - migration HTTP to HTTPS\\n?Site 66 - FW bypass E2E\\nVRRP improvements (LAB1 / L52)",
                "sites": [
                    "35COLO",
                    "35HEADOFFICE",
                    "35PRODUCTION",
                    "35SALES",
                    "38 Pilsen DR",
                    "66 Ostrava DC",
                    "68 Pardubice Distribution",
                    "ACI",
                    "AWS_SITE",
                    "AZURE",
                    "HWLAB",
                    "L31",
                    "L33",
                    "L34",
                    "L35",
                    "L36",
                    "L37",
                    "L39",
                    "L43",
                    "L45",
                    "L46",
                    "L47",
                    "L48",
                    "L49",
                    "L51",
                    "L52",
                    "L62",
                    "L63",
                    "L64",
                    "L65",
                    "L67",
                    "L71",
                    "L72",
                    "L77",
                    "L81",
                    "LAB01",
                    "MERAKI_SITE",
                    "MPLS",
                    "NSX-T",
                    "SILVERPEAK",
                    "VERSA_SITE",
                    "VIPTELA",
                ],
                "start": "2021-07-02T06:00:00.930000Z",
                "change": "2022-03-25T14:35:48.277000Z",
                "errors": [
                    {"count": 2, "error_type": "ABMapResultError"},
                    {"count": 5, "error_type": "ABParseError"},
                    {"count": 3, "error_type": "ABTaskMapResultError"},
                    {"count": 1, "error_type": "ABAmbiguousCommand"},
                    {"count": 7, "error_type": "ABCmdAuthFail"},
                    {"count": 6, "error_type": "ABCommandTimeout"},
                    {"count": 1, "error_type": "ABNoConfig"},
                    {"count": 1, "error_type": "ABParseBadConfigError"},
                    {"count": 1, "error_type": "ABTaskMapResultBadConfigError"},
                    {"count": 1, "error_type": "ABWorkerAuthError"},
                ],
                "locked": False,
                "status": "done",
                "loading": False,
                "version": "6.3.0-13",
                "user_count": 2324,
                "loaded_size": 170856074,
                "snapshot_id": "12dd8c61-129c-431a-b98b-4c9211571f89",
                "from_archive": True,
                "finish_status": "done",
                "unloaded_size": 26914884,
                "initial_version": "4.4.3+2",
                "interface_count": 9608,
                "total_dev_count": 729,
                "creator_username": None,
                "device_added_count": 0,
                "licensed_dev_count": 720,
                "device_removed_count": 0,
                "disabled_graph_cache": False,
                "interface_edge_count": 534,
                "interface_active_count": 6379,
                "disabled_historical_data": False,
                "disabled_intent_verification": False,
            },
            last_updated=timezone.now(),
        )
        sync = IPFabricSync.objects.create(
            name="ingest",
            type="dcim",
            status="new",
            snapshot_data=snapshot,
            parameters={
                "vrf": False,
                "site": True,
                "vlan": False,
                "sites": [],
                "device": True,
                "prefix": False,
                "platform": True,
                "interface": False,
                "ipaddress": False,
                "devicerole": True,
                "devicetype": True,
                "manufacturer": True,
                "virtualchassis": False,
            },
        )

        runner = IPFabricSyncRunner(
            transform_map=IPFabricTransformMap,
            settings={
                "site": True,
                "sites": [],
                "device": True,
                "platform": True,
                "interface": False,
                "devicerole": True,
                "devicetype": True,
                "manufacturer": True,
                "virtualchassis": True,
                "snapshot_id": "12dd8c61-129c-431a-b98b-4c9211571f89",
            },
            sync=sync,
        )
        device_uuid = str(uuid.uuid4())

        site_data = {
            "siteName": "Site 1",
            "devicesCount": 1,
            "usersCount": 2,
            "stpDCount": 0,
            "switchesCount": 0,
            "vlanCount": 1,
            "rDCount": 0,
            "routersCount": 0,
            "networksCount": 6,
        }

        self.site = runner.get_model_or_update(
            "dcim", "site", site_data, uuid=device_uuid
        )

        device_data = {
            "id": "961251111",
            "configReg": "0x0",
            "devType": "router",
            "family": "ios",
            "hostname": "L21PE152",
            "hostnameOriginal": None,
            "hostnameProcessed": None,
            "domain": None,
            "fqdn": None,
            "icon": None,
            "image": "unix:/opt/unetlab/addons/iol/bin/i86bi-linux-l3-adventerprisek9-15.2",
            "objectId": None,
            "taskKey": "fb67e3b4-5e48-4e52-b000-56cb187f2852",
            "loginIp": "10.21.254.152",
            "loginType": "telnet",
            "loginPort": None,
            "mac": None,
            "memoryTotalBytes": 396008048,
            "memoryUsedBytes": 72264172,
            "memoryUtilization": 18.25,
            "model": "",
            "platform": "i86bi_linux",
            "processor": "Intel-x86",
            "rd": "3",
            "reload": "reload at 0",
            "siteName": "MPLS",
            "sn": "a15ff98",
            "snHw": "a15ff98",
            "stpDomain": None,
            "uptime": 7254180,
            "vendor": "cisco",
            "version": "15.2(4)M1",
            "slug": None,
        }

        self.mf_obj = runner.get_model_or_update(
            "dcim", "manufacturer", device_data, uuid=device_uuid
        )
        self.dt_obj = runner.get_model_or_update(
            "dcim", "devicetype", device_data, uuid=device_uuid
        )

        self.platform = runner.get_model_or_update(
            "dcim", "platform", device_data, uuid=device_uuid
        )

        self.device_role = runner.get_model_or_update(
            "dcim", "devicerole", device_data, uuid=device_uuid
        )

        self.device_object = runner.get_model_or_update(
            "dcim", "device", device_data, uuid=device_uuid
        )

    def test_transform_map(self):
        site_transform_map = IPFabricTransformMap.objects.get(name="Site Transform Map")
        self.assertEqual(site_transform_map.name, "Site Transform Map")
        self.assertEqual(site_transform_map.source_model, "site")
        self.assertEqual(site_transform_map.status, "active")
        self.assertEqual(
            site_transform_map.target_model,
            ContentType.objects.filter(app_label="dcim", model="site")[0],
        )

    def test_transform_field(self):
        site_transform_map = IPFabricTransformMap.objects.get(name="Site Transform Map")
        site_slug_field = IPFabricTransformField.objects.get(
            source_field="siteName",
            target_field="slug",
            transform_map=site_transform_map,
        )
        self.assertEqual(site_slug_field.source_field, "siteName")
        self.assertEqual(site_slug_field.target_field, "slug")
        self.assertEqual(site_slug_field.template, "{{ object.siteName | slugify }}")
        self.assertEqual(site_slug_field.transform_map, site_transform_map)
        site_name_field = IPFabricTransformField.objects.get(
            source_field="siteName",
            target_field="name",
            transform_map=site_transform_map,
        )
        self.assertEqual(site_name_field.source_field, "siteName")
        self.assertEqual(site_name_field.target_field, "name")
        self.assertEqual(site_name_field.template, "")
        self.assertEqual(site_name_field.transform_map, site_transform_map)

    def test_transform_map_serialization(self):
        site_transform_map = IPFabricTransformMap.objects.get(name="Site Transform Map")
        data = serializers.serialize("json", [site_transform_map])
        data = json.loads(data)[0]
        test_data = {
            "model": "ipfabric_netbox.ipfabrictransformmap",
            "pk": site_transform_map.pk,
            "fields": {
                "name": "Site Transform Map",
                "source_model": "site",
                "target_model": 56,
                "status": "active",
            },
        }
        new_data = copy.deepcopy(data)
        new_data.pop("fields")
        new_fields = {}
        for k in test_data["fields"]:
            new_fields[k] = data["fields"][k]
        new_data["fields"] = new_fields
        self.assertDictEqual(test_data, new_data)

    def test_transform_field_serialization(self):
        site_transform_map = IPFabricTransformMap.objects.get(name="Site Transform Map")
        site_slug_field = IPFabricTransformField.objects.get(
            source_field="siteName", target_field="slug"
        )
        data = serializers.serialize("json", [site_slug_field])
        data = json.loads(data)[0]
        test_data = {
            "model": "ipfabric_netbox.ipfabrictransformfield",
            "pk": site_slug_field.pk,
            "fields": {
                "source_field": "siteName",
                "target_field": "slug",
                "template": "{{ object.siteName | slugify }}",
                "transform_map": site_transform_map.pk,
            },
        }
        new_data = copy.deepcopy(data)
        new_data.pop("fields")
        new_data["fields"] = {k: data["fields"][k] for k in test_data["fields"]}
        self.assertDictEqual(test_data, new_data)

    def test_update_or_create_instance_site(self):
        site_transform_map = IPFabricTransformMap.objects.get(name="Site Transform Map")
        data = {
            "siteName": "Site 1",
            "devicesCount": 1,
            "usersCount": 2,
            "stpDCount": 0,
            "switchesCount": 0,
            "vlanCount": 1,
            "rDCount": 0,
            "routersCount": 0,
            "networksCount": 6,
        }
        object = site_transform_map.update_or_create_instance(data=data)
        self.assertEqual(object.name, "Site 1")
        self.assertEqual(object.slug, "site-1")

    def test_update_or_create_instance_device(self):
        device_object = Device.objects.first()

        self.assertEqual(device_object.name, "L21PE152")
        self.assertEqual(device_object.serial, "a15ff98")
        self.assertEqual(device_object.platform, self.platform)
        self.assertEqual(device_object.device_role, self.device_role)
        self.assertEqual(device_object.device_type, self.dt_obj)
        self.assertEqual(device_object.device_type.manufacturer, self.mf_obj)
        self.assertEqual(device_object.site, self.site)
        self.assertEqual(device_object.status, "active")

    def test_alter_transform_field_template(self):
        sync = IPFabricSync.objects.get(name="ingest")

        runner = IPFabricSyncRunner(
            transform_map=IPFabricTransformMap,
            settings={
                "site": True,
                "sites": [],
                "device": True,
                "platform": True,
                "interface": False,
                "devicerole": True,
                "devicetype": True,
                "manufacturer": True,
                "virtualchassis": True,
                "snapshot_id": "12dd8c61-129c-431a-b98b-4c9211571f89",
            },
            sync=sync,
        )
        device_uuid = str(uuid.uuid4())

        device_data = {
            "id": "961251111",
            "configReg": "0x0",
            "devType": "router",
            "family": "ios",
            "hostname": "L21PE152",
            "hostnameOriginal": None,
            "hostnameProcessed": None,
            "domain": None,
            "fqdn": None,
            "icon": None,
            "image": "unix:/opt/unetlab/addons/iol/bin/i86bi-linux-l3-adventerprisek9-15.2",
            "objectId": None,
            "taskKey": "fb67e3b4-5e48-4e52-b000-56cb187f2852",
            "loginIp": "10.21.254.152",
            "loginType": "telnet",
            "loginPort": None,
            "mac": None,
            "memoryTotalBytes": 396008048,
            "memoryUsedBytes": 72264172,
            "memoryUtilization": 18.25,
            "model": "",
            "platform": "i86bi_linux",
            "processor": "Intel-x86",
            "rd": "3",
            "reload": "reload at 0",
            "siteName": "MPLS",
            "sn": "a15ff98",
            "snHw": "a15ff98",
            "stpDomain": None,
            "uptime": 7254180,
            "vendor": "cisco",
            "version": "15.2(4)M1",
            "slug": None,
        }

        transform_field = IPFabricTransformField.objects.get(pk=10)
        transform_field.template = "{{ object.hostname }} - test"
        transform_field.save()
        device_object = runner.get_model_or_update(
            "dcim", "device", device_data, uuid=device_uuid
        )
        self.assertEqual(device_object.name, "L21PE152 - test")



from typing import Optional

from dcim.models import Device
from dcim.models import Site
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from extras.choices import CustomFieldTypeChoices
from extras.choices import CustomLinkButtonClassChoices
from extras.models import CustomField
from extras.models import CustomLink
from packaging import version

from .models import IPFabricBranch
from .models import IPFabricSource

NETBOX_CURRENT_VERSION = version.parse(settings.VERSION)

if NETBOX_CURRENT_VERSION >= version.parse("3.7.0"):
    from extras.choices import (
        CustomFieldUIVisibleChoices as CustomFieldUIVisibleChoices,
    )
    from extras.choices import CustomFieldUIEditableChoices
else:
    from extras.choices import (
        CustomFieldVisibilityChoices as CustomFieldUIVisibleChoices,
    )


def create_custom_field(
    field_name: str,
    label: str,
    models: list,
    object_type=None,
    cf_type: Optional[str] = "type_text",
    ui_visibility: Optional[str] = "VISIBILITY_READ_ONLY",
):
    if NETBOX_CURRENT_VERSION >= version.parse("3.7.0"):
        defaults = {
            "label": label,
            "object_type": (
                ContentType.objects.get_for_model(object_type) if object_type else None
            ),
            "ui_visible": getattr(CustomFieldUIVisibleChoices, "ALWAYS"),
            "ui_editable": getattr(CustomFieldUIEditableChoices, "NO"),
        }
    else:
        defaults = {
            "label": label,
            "object_type": (
                ContentType.objects.get_for_model(object_type) if object_type else None
            ),
            "ui_visibility": getattr(
                CustomFieldUIVisibleChoices, "VISIBILITY_READ_ONLY"
            ),
        }

    custom_field, _ = CustomField.objects.update_or_create(
        type=getattr(CustomFieldTypeChoices, cf_type.upper()),
        name=field_name,
        defaults=defaults,
    )

    for model in models:
        custom_field.content_types.add(ContentType.objects.get_for_model(model))


def ipfabric_netbox_init():
    create_custom_field(
        "ipfabric_source",
        "IP Fabric Source",
        [Device, Site],
        cf_type="type_object",
        object_type=IPFabricSource,
    )
    create_custom_field(
        "ipfabric_branch",
        "IP Fabric Last Sync",
        [Device, Site],
        cf_type="type_object",
        object_type=IPFabricBranch,
    )
    cl, _ = CustomLink.objects.update_or_create(
        defaults={
            "link_text": "{% if object.custom_field_data.ipfabric_source is defined %}{% set SOURCE_ID = object.custom_field_data.ipfabric_source %}{% if SOURCE_ID %}IP Fabric{% endif %}{% endif %}",
            "link_url": '{% if object.custom_field_data.ipfabric_source is defined %}{% set SOURCE_ID = object.custom_field_data.ipfabric_source %}{% if SOURCE_ID %}{% set BASE_URL = object.custom_fields.filter(object_type__model="ipfabricsource").first().object_type.model_class().objects.get(pk=SOURCE_ID).url %}{{ BASE_URL }}/inventory/devices?options={"filters":{"sn": ["like","{{ object.serial }}"]}}{% endif %}{%endif%}',
            "new_window": True,
            "button_class": CustomLinkButtonClassChoices.BLUE,
        },
        name="ipfabric",
    )
    cl.content_types.add(ContentType.objects.get_for_model(Device))



# api/urls.py
from netbox.api.routers import NetBoxRouter

from ipfabric_netbox.api.views import IPFabricBranchViewSet
from ipfabric_netbox.api.views import IPFabricRelationshipFieldiewSet
from ipfabric_netbox.api.views import IPFabricSnapshotViewSet
from ipfabric_netbox.api.views import IPFabricSourceViewSet
from ipfabric_netbox.api.views import IPFabricSyncViewSet
from ipfabric_netbox.api.views import IPFabricTransformFieldiewSet
from ipfabric_netbox.api.views import IPFabricTransformMapViewSet


router = NetBoxRouter()
router.register("source", IPFabricSourceViewSet)
router.register("snapshot", IPFabricSnapshotViewSet)
router.register("transform-map", IPFabricTransformMapViewSet)
router.register("sync", IPFabricSyncViewSet)
router.register("branch", IPFabricBranchViewSet)
router.register("transform-field", IPFabricTransformFieldiewSet)
router.register("relationship-field", IPFabricRelationshipFieldiewSet)
urlpatterns = router.urls



from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import extend_schema_field
from netbox.api.fields import ContentTypeField
from netbox.api.serializers import NetBoxModelSerializer
from netbox.api.serializers import WritableNestedSerializer
from rest_framework import serializers
from users.api.nested_serializers import NestedUserSerializer

from ipfabric_netbox.models import IPFabricBranch
from ipfabric_netbox.models import IPFabricSnapshot
from ipfabric_netbox.models import IPFabricSource
from ipfabric_netbox.models import IPFabricSync
from ipfabric_netbox.models import IPFabricTransformMap

# from .serializers import IPFabricSyncSerializer

__all__ = (
    "NestedIPFabricSourceSerializer",
    "NestedIPFabricSnapshotSerializer",
    "NestedIPFabricTransformMapSerializer",
    "NestedIPFabricBranchSerializer",
    "NestedIPFabricSyncSerializer",
)


class NestedIPFabricSourceSerializer(WritableNestedSerializer):
    class Meta:
        model = IPFabricSource
        fields = ["id", "url", "display", "name", "type"]


class NestedIPFabricSnapshotSerializer(NetBoxModelSerializer):
    source = NestedIPFabricSourceSerializer(read_only=True)
    display = serializers.SerializerMethodField(read_only=True)

    class Meta:
        model = IPFabricSnapshot
        fields = [
            "id",
            "name",
            "source",
            "snapshot_id",
            "status",
            "date",
            "display",
            "sites",
        ]

    @extend_schema_field(OpenApiTypes.STR)
    def get_display(self, obj):
        return f"{obj.name} ({obj.snapshot_id})"


class NestedIPFabricSyncSerializer(NetBoxModelSerializer):
    snapshot_data = NestedIPFabricSnapshotSerializer(read_only=True)

    class Meta:
        model = IPFabricSync
        fields = [
            "id",
            "name",
            "display",
            "snapshot_data",
            "type",
            "status",
            "parameters",
            "last_synced",
        ]


class NestedIPFabricTransformMapSerializer(NetBoxModelSerializer):
    target_model = ContentTypeField(read_only=True)

    class Meta:
        model = IPFabricTransformMap
        fields = [
            "id",
            "source_model",
            "target_model",
            "status",
        ]


class NestedIPFabricBranchSerializer(NetBoxModelSerializer):
    user = NestedUserSerializer(read_only=True)
    sync = NestedIPFabricSyncSerializer(read_only=True)

    class Meta:
        model = IPFabricBranch
        fields = [
            "id",
            "name",
            "display",
            "sync",
            "description",
            "user",
        ]



# Generated by Django 4.1.10 on 2023-10-10 15:47
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("ipfabric_netbox", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="ipfabricsnapshot",
            name="status",
            field=models.CharField(default="unloaded", max_length=50),
        ),
    ]



import requests
from dcim.models import Device
from dcim.models import Site
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.cache import cache
from django.db import models
from django.shortcuts import get_object_or_404
from django.shortcuts import redirect
from django.shortcuts import render
from django.urls import reverse
from django.utils import timezone
from django.views.generic import View
from django_tables2 import RequestConfig
from extras.choices import ChangeActionChoices
from ipfabric.diagrams import Network
from ipfabric.diagrams import NetworkSettings
from netbox.staging import StagedChange
from netbox.views import generic
from netbox.views.generic.base import BaseObjectView
from utilities.forms import ConfirmationForm
from utilities.htmx import is_htmx
from utilities.paginator import EnhancedPaginator
from utilities.paginator import get_paginate_count
from utilities.utils import count_related
from utilities.utils import get_viewname
from utilities.utils import serialize_object
from utilities.utils import shallow_compare_dict
from utilities.views import register_model_view
from utilities.views import ViewTab

from .filtersets import IPFabricBranchFilterSet
from .filtersets import IPFabricDataFilterSet
from .filtersets import IPFabricSnapshotFilterSet
from .filtersets import IPFabricSourceFilterSet
from .filtersets import IPFabricStagedChangeFilterSet
from .forms import IPFabricBranchFilterForm
from .forms import IPFabricRelationshipFieldForm
from .forms import IPFabricSnapshotFilterForm
from .forms import IPFabricSourceFilterForm
from .forms import IPFabricSourceForm
from .forms import IPFabricSyncForm
from .forms import IPFabricTableForm
from .forms import IPFabricTransformFieldForm
from .forms import IPFabricTransformMapForm
from .models import IPFabricBranch
from .models import IPFabricData
from .models import IPFabricRelationshipField
from .models import IPFabricSnapshot
from .models import IPFabricSource
from .models import IPFabricSync
from .models import IPFabricTransformField
from .models import IPFabricTransformMap
from .tables import BranchTable
from .tables import DeviceIPFTable
from .tables import IPFabricDataTable
from .tables import IPFabricRelationshipFieldTable
from .tables import IPFabricSnapshotTable
from .tables import IPFabricSourceTable
from .tables import IPFabricTransformFieldTable
from .tables import IPFabricTransformMapTable
from .tables import StagedChangesTable
from .utilities.ipfutils import IPFabric
from .utilities.transform_map import BuildTransformMaps


# Transform Map Relationship Field


@register_model_view(IPFabricRelationshipField, "edit")
class IPFabricRelationshipFieldEditView(generic.ObjectEditView):
    queryset = IPFabricRelationshipField.objects.all()
    form = IPFabricRelationshipFieldForm
    default_return_url = "plugins:ipfabric_netbox:ipfabricrelationshipfield_list"


class IPFabricRelationshipFieldDeleteView(generic.ObjectDeleteView):
    queryset = IPFabricRelationshipField.objects.all()
    default_return_url = "plugins:ipfabric_netbox:ipfabricrelationshipfield_list"


@register_model_view(IPFabricTransformMap, "relationships")
class IPFabricTransformRelationshipView(generic.ObjectChildrenView):
    queryset = IPFabricTransformMap.objects.all()
    child_model = IPFabricRelationshipField
    table = IPFabricRelationshipFieldTable
    template_name = "ipfabric_netbox/inc/transform_map_relationship_map.html"
    tab = ViewTab(
        label="Relationship Maps",
        badge=lambda obj: IPFabricRelationshipField.objects.filter(
            transform_map=obj
        ).count(),
        permission="ipfabric_netbox.view_ipfabricrelationshipfield",
    )

    def get_children(self, request, parent):
        return self.child_model.objects.filter(transform_map=parent)


class IPFabricRelationshipFieldListView(generic.ObjectListView):
    queryset = IPFabricRelationshipField.objects.all()
    table = IPFabricRelationshipFieldTable


# Transform Map


class IPFabricTransformMapListView(generic.ObjectListView):
    queryset = IPFabricTransformMap.objects.all()
    table = IPFabricTransformMapTable
    template_name = "ipfabric_netbox/ipfabrictransformmap_list.html"


class IPFabricTransformMapRestoreView(generic.ObjectListView):
    queryset = IPFabricTransformMap.objects.all()
    table = IPFabricTransformMapTable

    def get_required_permission(self):
        return "ipfabric_netbox.tm_restore"

    def get(self, request):
        if is_htmx(request):
            viewname = get_viewname(self.queryset.model, action="restore")
            form_url = reverse(viewname)
            form = ConfirmationForm(initial=request.GET)
            dependent_objects = {
                IPFabricTransformMap: IPFabricTransformMap.objects.all(),
                IPFabricTransformField: IPFabricTransformField.objects.all(),
                IPFabricRelationshipField: IPFabricRelationshipField.objects.all(),
            }
            print(dependent_objects)
            return render(
                request,
                "ipfabric_netbox/ipfabrictransformmap_restore.html",
                {
                    "form": form,
                    "form_url": form_url,
                    "dependent_objects": dependent_objects,
                },
            )

    def post(self, request):
        IPFabricTransformMap.objects.all().delete()
        try:
            data = requests.get(
                "https://gitlab.com/ip-fabric/integrations/ipfabric-netbox/-/raw/main/scripts/transform_map.json"
            ).json()
        except Exception as e:
            messages.error(request, e)
        BuildTransformMaps(data=data)
        return redirect("plugins:ipfabric_netbox:ipfabrictransformmap_list")


@register_model_view(IPFabricTransformMap, "edit")
class IPFabricTransformMapEditView(generic.ObjectEditView):
    queryset = IPFabricTransformMap.objects.all()
    form = IPFabricTransformMapForm
    default_return_url = "plugins:ipfabric_netbox:ipfabrictransformmap_list"


class IPFabricTransformMapDeleteView(generic.ObjectDeleteView):
    queryset = IPFabricTransformMap.objects.all()
    default_return_url = "plugins:ipfabric_netbox:ipfabrictransformmap_list"


class IPFabricTransformMapBulkDeleteView(generic.BulkDeleteView):
    queryset = IPFabricTransformMap.objects.all()
    table = IPFabricTransformMapTable


@register_model_view(IPFabricTransformMap)
class IPFabricTransformMapView(generic.ObjectView):
    queryset = IPFabricTransformMap.objects.all()


# Transform Map Field


class IPFabricTransformFieldListView(generic.ObjectListView):
    queryset = IPFabricTransformField.objects.all()
    table = IPFabricTransformFieldTable


@register_model_view(IPFabricTransformField, "edit")
class IPFabricTransformFieldEditView(generic.ObjectEditView):
    queryset = IPFabricTransformField.objects.all()
    form = IPFabricTransformFieldForm


class IPFabricTransformFieldDeleteView(generic.ObjectDeleteView):
    queryset = IPFabricTransformField.objects.all()


@register_model_view(IPFabricTransformMap, "fields")
class IPFabricTransformFieldView(generic.ObjectChildrenView):
    queryset = IPFabricTransformMap.objects.all()
    child_model = IPFabricTransformField
    table = IPFabricTransformFieldTable
    template_name = "ipfabric_netbox/inc/transform_map_field_map.html"
    tab = ViewTab(
        label="Field Maps",
        badge=lambda obj: IPFabricTransformField.objects.filter(
            transform_map=obj
        ).count(),
        permission="ipfabric_netbox.view_ipfabrictransformfield",
    )

    def get_children(self, request, parent):
        return self.child_model.objects.filter(transform_map=parent)


# Snapshot


class IPFabricSnapshotListView(generic.ObjectListView):
    queryset = IPFabricSnapshot.objects.all()
    table = IPFabricSnapshotTable
    filterset = IPFabricSnapshotFilterSet
    filterset_form = IPFabricSnapshotFilterForm


@register_model_view(IPFabricSnapshot)
class IPFabricSnapshotView(generic.ObjectView):
    queryset = IPFabricSnapshot.objects.all()


class IPFabricSnapshotDeleteView(generic.ObjectDeleteView):
    queryset = IPFabricSnapshot.objects.all()


class IPFabricSnapshotBulkDeleteView(generic.BulkDeleteView):
    queryset = IPFabricSnapshot.objects.all()
    filterset = IPFabricSnapshotFilterSet
    table = IPFabricSnapshotTable


@register_model_view(IPFabricSnapshot, "data")
class IPFabricSnapshotRawView(generic.ObjectChildrenView):
    queryset = IPFabricSnapshot.objects.all()
    child_model = IPFabricData
    table = IPFabricDataTable
    template_name = "ipfabric_netbox/inc/snapshotdata.html"
    tab = ViewTab(
        label="Raw Data",
        badge=lambda obj: IPFabricData.objects.filter(snapshot_data=obj).count(),
        permission="ipfabric_netbox.view_ipfabricsnapshot",
        hide_if_empty=True,
    )

    def get_children(self, request, parent):
        return self.child_model.objects.filter(snapshot_data=parent)


class IPFabricSnapshotDataDeleteView(generic.ObjectDeleteView):
    queryset = IPFabricData.objects.all()


class IPFabricSnapshotDataBulkDeleteView(generic.BulkDeleteView):
    queryset = IPFabricData.objects.all()
    filterset = IPFabricDataFilterSet
    table = IPFabricDataTable


@register_model_view(
    IPFabricData,
    name="data",
    path="json",
    kwargs={},
)
class IPFabricSnapshotDataJSONView(LoginRequiredMixin, View):
    template_name = "ipfabric_netbox/inc/json.html"

    def get(self, request, **kwargs):
        print(kwargs)
        # change_id = kwargs.get("change_pk", None)

        if is_htmx(request):
            data = get_object_or_404(IPFabricData, pk=kwargs.get("pk"))
            return render(
                request,
                self.template_name,
                {
                    "object": data,
                },
            )

        #         return render(
        #             request,
        #             self.template_name,
        #             {
        #                 "change": change,
        #                 "prechange_data": prechange_data,
        #                 "postchange_data": postchange_data,
        #                 "diff_added": diff_added,
        #                 "diff_removed": diff_removed,
        #                 "size": "lg",
        #             },
        #         )


# Source


class IPFabricSourceListView(generic.ObjectListView):
    queryset = IPFabricSource.objects.annotate(
        snapshot_count=count_related(IPFabricSnapshot, "source")
    )
    filterset = IPFabricSourceFilterSet
    filterset_form = IPFabricSourceFilterForm
    table = IPFabricSourceTable


@register_model_view(IPFabricSource, "edit")
class IPFabricSourceEditView(generic.ObjectEditView):
    queryset = IPFabricSource.objects.all()
    form = IPFabricSourceForm


@register_model_view(IPFabricSource)
class IPFabricSourceView(generic.ObjectView):
    queryset = IPFabricSource.objects.all()

    def get_extra_context(self, request, instance):
        related_models = (
            (
                IPFabricSnapshot.objects.restrict(request.user, "view").filter(
                    source=instance
                ),
                "source_id",
            ),
        )

        job = instance.jobs.order_by("id").last()
        data = {"related_models": related_models, "job": job}
        if job:
            data["job_results"] = job.data
        return data


@register_model_view(IPFabricSource, "sync")
class IPFabricSourceSyncView(BaseObjectView):
    queryset = IPFabricSource.objects.all()

    def get_required_permission(self):
        return "ipfabric_netbox.sync_source"

    def get(self, request, pk):
        ipfabricsource = get_object_or_404(self.queryset, pk=pk)
        return redirect(ipfabricsource.get_absolute_url())

    def post(self, request, pk):
        ipfabricsource = get_object_or_404(self.queryset, pk=pk)
        job = ipfabricsource.enqueue_sync_job(request=request)

        messages.success(request, f"Queued job #{job.pk} to sync {ipfabricsource}")
        return redirect(ipfabricsource.get_absolute_url())


@register_model_view(IPFabricSource, "delete")
class IPFabricSourceDeleteView(generic.ObjectDeleteView):
    queryset = IPFabricSource.objects.all()


class IPFabricSourceBulkDeleteView(generic.BulkDeleteView):
    queryset = IPFabricSource.objects.all()
    filterset = IPFabricSourceFilterSet
    table = IPFabricSourceTable


# Sync


class IPFabricSyncListView(View):
    def get(self, request):
        syncs = IPFabricSync.objects.prefetch_related("snapshot_data")
        return render(
            request,
            "ipfabric_netbox/sync_list.html",
            {"model": IPFabricSync, "syncs": syncs},
        )


@register_model_view(IPFabricSync, "edit")
class IPFabricSyncEditView(generic.ObjectEditView):
    queryset = IPFabricSync.objects.all()
    form = IPFabricSyncForm

    def alter_object(self, obj, request, url_args, url_kwargs):
        obj.user = request.user
        return obj


@register_model_view(IPFabricSync)
class IPFabricSyncView(generic.ObjectView):
    queryset = IPFabricSync.objects.all()
    actions = ("edit",)

    def get(self, request, **kwargs):
        instance = self.get_object(**kwargs)
        last_branch = instance.ipfabricbranch_set.last()

        if is_htmx(request):
            response = render(
                request,
                "ipfabric_netbox/partials/sync_last_branch.html",
                {"last_branch": last_branch},
            )

            if instance.status not in ["queued", "syncing"]:
                messages.success(
                    request,
                    f"Ingestion ({instance.name}) {instance.status}. Branch {last_branch.name} {last_branch.job.status}.",
                )
                response["HX-Refresh"] = "true"
            return response

        return render(
            request,
            self.get_template_name(),
            {
                "object": instance,
                "tab": self.tab,
                **self.get_extra_context(request, instance),
            },
        )

    def get_extra_context(self, request, instance):
        last_branch = instance.ipfabricbranch_set.last()

        if request.GET.get("format") in ["json", "yaml"]:
            format = request.GET.get("format")
            if request.user.is_authenticated:
                request.user.config.set("data_format", format, commit=True)
        elif request.user.is_authenticated:
            format = request.user.config.get("data_format", "json")
        else:
            format = "json"

        last_branch = instance.ipfabricbranch_set.last()

        return {"format": format, "last_branch": last_branch}


@register_model_view(IPFabricSync, "delete")
class IPFabricSyncDeleteView(generic.ObjectDeleteView):
    queryset = IPFabricSync.objects.all()
    default_return_url = "plugins:ipfabric_netbox:ipfabricsync_list"

    def get(self, request, pk):
        obj = get_object_or_404(self.queryset, pk=pk)

        if is_htmx(request):
            viewname = get_viewname(self.queryset.model, action="delete")
            form_url = reverse(viewname, kwargs={"pk": obj.pk})
            form = ConfirmationForm(initial=request.GET)
            return render(
                request,
                "ipfabric_netbox/inc/sync_delete.html",
                {
                    "object": obj,
                    "object_type": self.queryset.model._meta.verbose_name,
                    "form": form,
                    "form_url": form_url,
                    **self.get_extra_context(request, obj),
                },
            )


class IPFabricSyncBulkDeleteView(generic.BulkDeleteView):
    queryset = IPFabricSync.objects.all()
    filterset = IPFabricSnapshotFilterSet
    table = IPFabricSnapshotTable


# Ingestion


@register_model_view(IPFabricSync, "sync")
class IPFabricIngestSyncView(BaseObjectView):
    queryset = IPFabricSync.objects.all()

    def get_required_permission(self):
        return "ipfabric_netbox.sync_ingest"

    def get(self, request, pk):
        ipfabric = get_object_or_404(self.queryset, pk=pk)
        return redirect(ipfabric.get_absolute_url())

    def post(self, request, pk):
        ipfabric = get_object_or_404(self.queryset, pk=pk)
        job = ipfabric.enqueue_sync_job(user=request.user, adhoc=True)

        messages.success(request, f"Queued job #{job.pk} to sync {ipfabric}")
        return redirect(ipfabric.get_absolute_url())


# Branch


class IPFabricBranchListView(generic.ObjectListView):
    queryset = IPFabricBranch.objects.all()
    filterset = IPFabricBranchFilterSet
    filterset_form = IPFabricBranchFilterForm
    table = BranchTable


@register_model_view(
    IPFabricBranch,
    name="logs",
    path="logs",
)
class IPFabricBranchLogView(LoginRequiredMixin, View):
    template_name = "ipfabric_netbox/partials/branch_all.html"

    def get(self, request, **kwargs):
        branch_id = kwargs.get("pk")
        if is_htmx(request):
            branch = IPFabricBranch.objects.get(pk=branch_id)
            data = branch.get_statistics()
            data["object"] = branch
            data["job"] = branch.job
            response = render(
                request,
                self.template_name,
                data,
            )
            if branch.job.completed:
                response["HX-Refresh"] = "true"
                return response
            else:
                return response


@register_model_view(IPFabricBranch)
class IPFabricBranchView(generic.ObjectView):
    queryset = IPFabricBranch.objects.annotate(
        num_created=models.Count(
            "staged_changes",
            filter=models.Q(staged_changes__action=ChangeActionChoices.ACTION_CREATE)
            & ~models.Q(staged_changes__object_type__model="objectchange"),
        ),
        num_updated=models.Count(
            "staged_changes",
            filter=models.Q(staged_changes__action=ChangeActionChoices.ACTION_UPDATE)
            & ~models.Q(staged_changes__object_type__model="objectchange"),
        ),
        num_deleted=models.Count(
            "staged_changes",
            filter=models.Q(staged_changes__action=ChangeActionChoices.ACTION_DELETE)
            & ~models.Q(staged_changes__object_type__model="objectchange"),
        ),
    )

    def get_extra_context(self, request, instance):
        data = instance.get_statistics()
        return data


@register_model_view(IPFabricBranch, "merge")
class IPFabricBranchMergeView(BaseObjectView):
    queryset = IPFabricBranch.objects.all()
    template_name = "ipfabric_netbox/inc/merge_form.html"

    def get_required_permission(self):
        return "ipfabric_netbox.merge_branch"

    def get(self, request, pk):
        obj = get_object_or_404(self.queryset, pk=pk)

        if is_htmx(request):
            viewname = get_viewname(self.queryset.model, action="merge")
            form_url = reverse(viewname, kwargs={"pk": obj.pk})
            form = ConfirmationForm(initial=request.GET)
            return render(
                request,
                "ipfabric_netbox/inc/merge_form.html",
                {
                    "object": obj,
                    "object_type": self.queryset.model._meta.verbose_name,
                    "form": form,
                    "form_url": form_url,
                    **self.get_extra_context(request, obj),
                },
            )

        return redirect(obj.get_absolute_url())

    def post(self, request, pk):
        ipfabric = get_object_or_404(self.queryset, pk=pk)
        job = ipfabric.enqueue_merge_job(user=request.user)

        messages.success(request, f"Queued job #{job.pk} to sync {ipfabric}")
        return redirect(ipfabric.get_absolute_url())


@register_model_view(
    IPFabricBranch,
    name="change_diff",
    path="change/<int:change_pk>",
    kwargs={"model": IPFabricBranch},
)
class IPFabricBranchChangesDiffView(LoginRequiredMixin, View):
    template_name = "ipfabric_netbox/inc/diff.html"

    def get(self, request, model, **kwargs):
        change_id = kwargs.get("change_pk", None)

        if is_htmx(request):
            if change_id:
                change = StagedChange.objects.get(pk=change_id)
                if hasattr(change.object, "pk"):
                    prechange_data = serialize_object(change.object, resolve_tags=False)
                    prechange_data = dict(sorted(prechange_data.items()))
                else:
                    prechange_data = None

                if hasattr(change, "data"):
                    postchange_data = dict(sorted(change.data.items()))

                if prechange_data and postchange_data:
                    diff_added = shallow_compare_dict(
                        prechange_data or dict(),
                        postchange_data or dict(),
                        exclude=["last_updated"],
                    )
                    diff_removed = (
                        {x: prechange_data.get(x) for x in diff_added}
                        if prechange_data
                        else {}
                    )
                else:
                    diff_added = None
                    diff_removed = None

                return render(
                    request,
                    self.template_name,
                    {
                        "change": change,
                        "prechange_data": prechange_data,
                        "postchange_data": postchange_data,
                        "diff_added": diff_added,
                        "diff_removed": diff_removed,
                        "size": "lg",
                    },
                )


@register_model_view(IPFabricBranch, "change")
class IPFabricBranchChangesView(generic.ObjectChildrenView):
    queryset = IPFabricBranch.objects.all()
    child_model = StagedChange
    table = StagedChangesTable
    filterset = IPFabricStagedChangeFilterSet
    template_name = "generic/object_children.html"
    tab = ViewTab(
        label="Changes",
        badge=lambda obj: StagedChange.objects.filter(branch=obj).count(),
        permission="ipfabric_netbox.view_ipfabricbranch",
    )

    def get_children(self, request, parent):
        return self.child_model.objects.filter(branch=parent)


@register_model_view(IPFabricBranch, "delete")
class IPFabricBranchDeleteView(generic.ObjectDeleteView):
    queryset = IPFabricBranch.objects.all()


@register_model_view(IPFabricSync, "branch")
class IPFabricBranchTabView(generic.ObjectChildrenView):
    queryset = IPFabricSync.objects.all()
    child_model = IPFabricBranch
    table = BranchTable
    filterset = IPFabricBranchFilterSet
    template_name = "generic/object_children.html"
    tab = ViewTab(
        label="Branches",
        badge=lambda obj: IPFabricBranch.objects.filter(sync=obj).count(),
        permission="ipfabric_netbox.view_ipfabricbranch",
    )

    def get_children(self, request, parent):
        return self.child_model.objects.filter(sync=parent)


@register_model_view(Device, "ipfabric")
class IPFabricTable(View):
    template_name = "ipfabric_netbox/ipfabric_table.html"
    tab = ViewTab("IP Fabric", permission="ipfabric_netbox.view_devicetable")

    def get(self, request, pk):
        device = get_object_or_404(Device, pk=pk)
        form = (
            IPFabricTableForm(request.GET)
            if "table" in request.GET
            else IPFabricTableForm()
        )
        data = None

        if form.is_valid():
            table = form.cleaned_data["table"]
            test = {
                "True": True,
                "False": False,
            }
            cache_enable = test.get(form.cleaned_data["cache_enable"])
            snapshot_id = ""

            if not form.cleaned_data["snapshot_data"]:
                snapshot_id = "$last"
                source = IPFabricSource.objects.get(
                    pk=device.custom_field_data["ipfabric_source"]
                )

            else:
                snapshot_id = form.cleaned_data["snapshot_data"].snapshot_id
                source = form.cleaned_data["snapshot_data"].source

            source.parameters["snapshot_id"] = snapshot_id
            source.parameters["base_url"] = source.url

            cache_key = (
                f"ipfabric_{table}_{device.serial}_{source.parameters['snapshot_id']}"
            )
            if cache_enable:
                data = cache.get(cache_key)

            if not data:
                try:
                    ipf = IPFabric(parameters=source.parameters)
                    raw_data, columns = ipf.get_table_data(table=table, device=device)
                    data = {"data": raw_data, "columns": columns}
                    cache.set(cache_key, data, 60 * 60 * 24)
                except Exception as e:
                    messages.error(request, e)

        if not data:
            data = {"data": [], "columns": []}

        table = DeviceIPFTable(data["data"], extra_columns=data["columns"])

        RequestConfig(
            request,
            {
                "paginator_class": EnhancedPaginator,
                "per_page": get_paginate_count(request),
            },
        ).configure(table)

        if is_htmx(request):
            return render(
                request,
                "htmx/table.html",
                {
                    "table": table,
                },
            )

        source = None

        if source_id := device.custom_field_data["ipfabric_source"]:
            source = IPFabricSource.objects.get(pk=source_id)

        return render(
            request,
            self.template_name,
            {
                "object": device,
                "source": source,
                "tab": self.tab,
                "form": form,
                "table": table,
            },
        )


@register_model_view(
    IPFabricSource,
    name="topology",
    path="topology/<int:site>",
    kwargs={"snapshot": ""},
)
class IPFabricSourceTopology(LoginRequiredMixin, View):
    template_name = "ipfabric_netbox/inc/site_topology_modal.html"

    def get(self, request, pk, site, **kwargs):
        if is_htmx(request):

            try:
                site = get_object_or_404(Site, pk=site)
                source_id = request.GET.get("source")
                if not source_id:
                    raise Exception("Source ID not available in request.")
                source = get_object_or_404(IPFabricSource, pk=source_id)
                snapshot = request.GET.get("snapshot")
                if not snapshot:
                    raise Exception("Snapshot ID not available in request.")

                source.parameters.update(
                    {"snapshot_id": snapshot, "base_url": source.url}
                )

                ipf = IPFabric(parameters=source.parameters)
                snapshot_data = ipf.ipf.snapshots.get(snapshot)
                if not snapshot_data:
                    raise Exception(
                        f"Snapshot ({snapshot}) not available in IP Fabric."
                    )

                sites = ipf.ipf.inventory.sites.all(
                    filters={"siteName": ["eq", site.name]}
                )
                if not sites:
                    raise Exception(
                        f"{site.name} not available in snapshot ({snapshot})."
                    )

                net = Network(sites=site.name, all_network=False)
                settings = NetworkSettings()
                settings.hide_protocol("xdp")
                settings.hiddenDeviceTypes.extend(["transit", "cloud"])

                link = ipf.ipf.diagram.share_link(net, graph_settings=settings)
                svg_data = ipf.ipf.diagram.svg(net, graph_settings=settings).decode(
                    "utf-8"
                )
                error = None
            except Exception as e:
                error = e
                svg_data = link = None

            return render(
                request,
                self.template_name,
                {
                    "site": site,
                    "source": source,
                    "svg": svg_data,
                    "size": "xl",
                    "link": link,
                    "time": timezone.now(),
                    "snapshot": snapshot_data,
                    "error": error,
                },
            )



import requests
from django.db import connection
from extras.plugins import PluginConfig


class NetboxIPFabricConfig(PluginConfig):
    name = "ipfabric_netbox"
    verbose_name = "NetBox IP Fabric SoT Plugin"
    description = "Sync IP Fabric into NetBox"
    version = "2.0.5"
    base_url = "ipfabric"

    def ready(self):
        super().ready()
        try:
            from ipfabric_netbox.signals import ipfabric_netbox_init

            all_tables = connection.introspection.table_names()
            if "extras_customfield" in all_tables:
                ipfabric_netbox_init()
            if "ipfabric_netbox_ipfabrictransformmap" in all_tables:
                from ipfabric_netbox.models import IPFabricTransformMap
                from ipfabric_netbox.utilities.transform_map import BuildTransformMaps

                if IPFabricTransformMap.objects.count() == 0:
                    data = requests.get(
                        "https://gitlab.com/ip-fabric/integrations/ipfabric-netbox/-/raw/main/scripts/transform_map.json"
                    ).json()
                    BuildTransformMaps(data=data)
        except Exception as e:
            print(f"Failed to initialize IP Fabric plugin: {e}.")


config = NetboxIPFabricConfig



from netbox.plugins import PluginTemplateExtension


class SiteTopologyButtons(PluginTemplateExtension):
    model = "dcim.site"

    def buttons(self):
        return self.render(
            "ipfabric_netbox/inc/site_topology_button.html", extra_context={}
        )


template_extensions = [SiteTopologyButtons]



# Generated by Django 4.1.10 on 2024-01-09 16:01
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("ipfabric_netbox", "0003_ipfabricsource_type_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="ipfabricsync",
            name="auto_merge",
            field=models.BooleanField(default=False),
        ),
    ]



from django.utils.translation import gettext_lazy as _
from utilities.choices import ChoiceSet

transform_field_source_columns = {
    "site": [
        "id",
        "siteName",
        "devicesCount",
        "usersCount",
        "stpDCount",
        "switchesCount",
        "vlanCount",
        "rDCount",
        "routersCount",
        "networksCount",
    ],
    "device": [
        "id",
        "sn",
        "hostname",
        "siteName",
        "rd",
        "stpDomain",
        "snHw",
        "loginIp",
        "objectId",
        "loginType",
        "uptime",
        "reload",
        "memoryUtilization",
        "memoryTotalBytes",
        "memoryUsedBytes",
        "vendor",
        "family",
        "platform",
        "model",
        "configReg",
        "version",
        "image",
        "processor",
        "mac",
        "devType",
        "hostnameOriginal",
        "hostnameProcessed",
        "domain",
        "fqdn",
    ],
    "inventory": [
        "id",
        "deviceSn",
        "hostname",
        "siteName",
        "deviceId",
        "name",
        "dscr",
        "pid",
        "sn",
        "vid",
        "vendor",
        "platform",
        "model",
    ],
    "interface": [
        "id",
        "dscr",
        "duplex",
        "errDisabled",
        "hasTransceiver",
        "hostname",
        "intName",
        "intNameAlias",
        "l1",
        "l2",
        "loginIp",
        "loginType",
        "mac",
        "media",
        "mtu",
        "nameOriginal",
        "primaryIp",
        "reason",
        "rel",
        "siteName",
        "sn",
        "speed",
        "speedValue",
        "transceiverPn",
        "transceiverSn",
        "transceiverType",
    ],
    "part_number": [
        "id",
        "deviceSn",
        "hostname",
        "siteName",
        "deviceId",
        "name",
        "dscr",
        "pid",
        "sn",
        "vid",
        "vendor",
        "platform",
        "model",
    ],
    "vlan": ["id", "siteName", "vlanId", "vlanName", "dscr", "devCount"],
    "vrf": ["id", "sn", "hostname", "siteName", "vrf", "rd", "intCount"],
    "prefix": ["id", "siteName", "net", "hosts", "gw", "gwV", "vrf", "vlanId"],
    "virtualchassis": [
        "id",
        "sn",
        "master",
        "siteName",
        "uptime",
        "member",
        "connectionsCount",
        "pn",
        "memberSn",
        "role",
        "state",
        "mac",
        "ver",
        "image",
        "hwVer",
    ],
    "ipaddress": [
        "hostname",
        "sn",
        "intName",
        "stateL1",
        "stateL2",
        "siteName",
        "dnsName",
        "dnsHostnameMatch",
        "vlanId",
        "dnsReverseMatch",
        "mac",
        "ip",
        "net",
        "type",
        "vrf",
    ],
}


class IPFabricTransformMapSourceModelChoices(ChoiceSet):
    SITE = "site"
    INVENTORY = "inventory"
    DEVICE = "device"
    VIRTUALCHASSIS = "virtualchassis"
    INTERFACE = "interface"
    VLAN = "vlan"
    VRF = "vrf"
    PREFIX = "prefix"
    IPADDRESS = "ipaddress"
    # PARTNUMBERS = "part_number"

    CHOICES = (
        (SITE, "Site", "cyan"),
        (INVENTORY, "Inventory", "gray"),
        (DEVICE, "Device", "gray"),
        (VIRTUALCHASSIS, "Virtual Chassis", "grey"),
        (INTERFACE, "Interface", "gray"),
        (VLAN, "VLAN", "gray"),
        (VRF, "VRF", "gray"),
        (PREFIX, "Prefix", "gray"),
        (IPADDRESS, "IP Address", "gray"),
        # (PARTNUMBERS, "Part Number", "gray"),
    )


class IPFabricSnapshotStatusModelChoices(ChoiceSet):
    key = "IPFabricSnapshot.status"

    STATUS_LOADED = "loaded"
    STATUS_UNLOADED = "unloaded"

    CHOICES = [
        (STATUS_LOADED, _("Loaded"), "green"),
        (STATUS_UNLOADED, _("Unloaded"), "red"),
    ]


class IPFabricSyncTypeChoices(ChoiceSet):
    ALL = "all"
    DCIM = "dcim"
    IPAM = "ipam"

    CHOICES = (
        (ALL, _("All"), "gray"),
        (DCIM, _("DCIM"), "blue"),
        (IPAM, _("IPAM"), "blue"),
    )


class IPFabricSourceTypeChoices(ChoiceSet):
    LOCAL = "local"
    REMOTE = "remote"

    CHOICES = (
        (LOCAL, "Local", "cyan"),
        (REMOTE, "Remote", "gray"),
    )


class IPFabricRawDataTypeChoices(ChoiceSet):
    DEVICE = "device"
    VLAN = "vlan"
    VRF = "vrf"
    VIRTUALCHASSIS = "virtualchassis"
    PREFIX = "prefix"
    INTERFACE = "interface"
    IPADDRESS = "ipaddress"
    INVENTORYITEM = "inventoryitem"
    SITE = "site"

    CHOICES = (
        (DEVICE, "Local", "cyan"),
        (VLAN, "VLAN", "gray"),
        (VIRTUALCHASSIS, "Virtual Chassis", "gray"),
        (PREFIX, "Prefix", "gray"),
        (INTERFACE, "Interface", "gray"),
        (INVENTORYITEM, "Inventory Item", "gray"),
        (IPADDRESS, "IP Address", "gray"),
        (SITE, "Site", "gray"),
    )



import django_filters
from core.choices import DataSourceStatusChoices
from django.db.models import Q
from django.utils.translation import gettext as _
from extras.choices import ChangeActionChoices
from netbox.filtersets import BaseFilterSet
from netbox.filtersets import ChangeLoggedModelFilterSet
from netbox.filtersets import NetBoxModelFilterSet
from netbox.staging import StagedChange

from .models import IPFabricBranch
from .models import IPFabricData
from .models import IPFabricSnapshot
from .models import IPFabricSource
from .models import IPFabricSync
from .models import IPFabricTransformMap


class IPFabricStagedChangeFilterSet(BaseFilterSet):
    q = django_filters.CharFilter(method="search")
    action = django_filters.MultipleChoiceFilter(choices=ChangeActionChoices)

    class Meta:
        model = StagedChange
        fields = ["branch", "action", "object_type"]

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(data__values__contains=value)
            | Q(action__icontains=value)
            | Q(object_type__model__icontains=value)
        )


class IPFabricDataFilterSet(BaseFilterSet):
    q = django_filters.CharFilter(method="search")

    class Meta:
        model = IPFabricData
        fields = ["snapshot_data", "type"]

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(snapshot_data__icontains=value) | Q(type__icontains=value)
        )


class IPFabricSnapshotFilterSet(ChangeLoggedModelFilterSet):
    q = django_filters.CharFilter(method="search")
    source_id = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricSource.objects.all(),
        label=_("Source (ID)"),
    )
    source = django_filters.ModelMultipleChoiceFilter(
        field_name="source__name",
        queryset=IPFabricSource.objects.all(),
        to_field_name="name",
        label=_("Source (name)"),
    )
    snapshot_id = django_filters.CharFilter(
        label=_("Snapshot ID"), lookup_expr="icontains"
    )

    class Meta:
        model = IPFabricSnapshot
        fields = ("id", "name", "status", "snapshot_id")

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(Q(name__icontains=value))


class IPFabricSourceFilterSet(NetBoxModelFilterSet):
    status = django_filters.MultipleChoiceFilter(
        choices=DataSourceStatusChoices, null_value=None
    )

    class Meta:
        model = IPFabricSource
        fields = ("id", "name")

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(name__icontains=value)
            | Q(description__icontains=value)
            | Q(comments__icontains=value)
        )


class IPFabricBranchFilterSet(BaseFilterSet):
    q = django_filters.CharFilter(method="search")
    sync_id = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricSync.objects.all(),
        label=_("Sync (ID)"),
    )
    sync = django_filters.ModelMultipleChoiceFilter(
        field_name="sync__name",
        queryset=IPFabricSync.objects.all(),
        to_field_name="name",
        label=_("Sync (name)"),
    )

    class Meta:
        model = IPFabricBranch
        fields = ("id", "name", "sync")

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(name__icontains=value) | Q(sync__name__icontains=value)
        )


class IPFabricTransformFieldFilterSet(BaseFilterSet):
    transform_map = django_filters.ModelMultipleChoiceFilter(
        queryset=IPFabricTransformMap.objects.all(), label=_("Transform Map")
    )



from collections import Counter
from copy import deepcopy

from dcim.models import Device
from dcim.models import InventoryItem
from dcim.models import Manufacturer

DEFAULT_DEVICE_ROLE = "Network Device"
device_serial_max_length = Device._meta.get_field("serial").max_length


def order_members(members):
    devices = {}

    for member in members:
        master_serial = member.get("sn")
        if master_serial and member.get("memberSn"):
            if master_serial in devices:
                devices[master_serial].append(member)
            else:
                devices[master_serial] = [member]

    return devices


def order_devices(devices, members):
    hostnames = [d["hostname"] for d in devices]
    counter = Counter(hostnames)

    new_devices = []

    for device in devices:
        if counter[device["hostname"]] > 1:
            device["hostname"] = f"{device['hostname']} - ({device['sn']})"
        if child_members := members.get(device.get("sn")):
            for child_member in child_members:
                if device.get("sn") != child_member.get("memberSn"):
                    new_device = deepcopy(device)
                    new_device["hostname"] = (
                        f"{device['hostname']}/{child_member.get('member')}"
                    )
                    new_device["model"] = child_member.get("pn")
                    new_device["sn"] = child_member.get("memberSn")
                    new_device["virtual_chassis"] = child_member
                    new_devices.append(new_device)
                else:
                    device["virtual_chassis"] = child_member
            hostnames = [d["hostname"] for d in devices]
            counter = Counter(hostnames)

    devices.extend(new_devices)

    return devices


def order_vrf(vrfs):
    vrf_dict = {}

    for vrf in vrfs:
        sn = vrf["sn"]
        if sn in vrf_dict:
            vrf_dict[sn].append(vrf)
        else:
            vrf_dict[sn] = [vrf]

    return vrf_dict


def order_pn(part_numbers):
    pn_dict = {}

    for pn in part_numbers:
        device_sn = pn["deviceSn"]
        if device_sn in pn_dict:
            pn_dict[device_sn].append(pn)
        else:
            pn_dict[device_sn] = [pn]

    return pn_dict


def create_inventory_items(device: Device, parts: list, manufacturer: Manufacturer):
    for part in parts:
        name = part.get("name", "")
        if len(name) > InventoryItem._meta.get_field("name").max_length:
            if part.get("dscr"):
                name = part.get("dscr")
            else:
                name = part.get("sn")

        defaults = {
            "name": name,
            "manufacturer": manufacturer,
            "serial": part.get("sn", ""),
            # "description": part.get('dscr', "123"),
            "part_id": part.get("pid", ""),
            "device": device,
            "lft": device.pk,
        }
        if part.get("dscr"):
            defaults["description"] = part.get("dscr")

        inventory_object, _ = InventoryItem.objects.update_or_create(
            serial=part.get("sn", ""), defaults=defaults
        )
